from robot.api.parsing import ModelTransformer, Tags, Token, DefaultTags, ForceTags


class OrderTags(ModelTransformer):
    """
    Order tags.

    Tags are ordered in lexicographic order like this:

        *** Test Cases ***
        Tags Upper Lower
            [Tags]    ba    Ab    Bb    Ca    Cb    aa
            My Keyword

        *** Keywords ***
        My Keyword
            [Tags]    ba    Ab    Bb    Ca    Cb    aa
            No Operation

    To:

        *** Test Cases ***
        Tags Upper Lower
            [Tags]    aa    Ab    ba    Bb    Ca    Cb
            My Keyword

        *** Keywords ***
        My Keyword
            [Tags]    aa    Ab    ba    Bb    Ca    Cb
            No Operation

    Default order can be changed using following parameters:
      - ``case_sensitive = False``
      - ``reverse = False``

    See https://robotidy.readthedocs.io/en/latest/transformers/OrderTags.html for more examples.
    """

    ENABLED = False

    def __init__(
        self,
        case_sensitive: bool = False,
        reverse: bool = False,
        default_tags: bool = True,
        force_tags: bool = True,
    ):
        self.key = self.get_key(case_sensitive)
        self.reverse = reverse
        self.default_tags = default_tags
        self.force_tags = force_tags

    def visit_Tags(self, node):  # noqa
        return self.order_tags(node, Tags, indent=True)

    def visit_DefaultTags(self, node):  # noqa
        return self.order_tags(node, DefaultTags) if self.default_tags else node

    def visit_ForceTags(self, node):  # noqa
        return self.order_tags(node, ForceTags) if self.force_tags else node

    def order_tags(self, node, tag_class, indent=False):
        ordered_tags = sorted(
            (tag.value for tag in node.data_tokens[1:]),
            key=self.key,
            reverse=self.reverse,
        )
        if len(ordered_tags) <= 1:
            return node
        comments = node.get_tokens(Token.COMMENT)
        if indent:
            tag_node = tag_class.from_params(
                ordered_tags,
                indent=self.formatting_config.separator,
                separator=self.formatting_config.separator,
            )
        else:
            tag_node = tag_class.from_params(ordered_tags, separator=self.formatting_config.separator)
        if comments:
            tag_node.tokens = tag_node.tokens[:-1] + tuple(self.join_tokens(comments)) + (tag_node.tokens[-1],)
        return tag_node

    def join_tokens(self, tokens):
        joined_tokens = []
        for token in tokens:
            joined_tokens.append(Token(Token.SEPARATOR, self.formatting_config.separator))
            joined_tokens.append(token)
        return joined_tokens

    @staticmethod
    def get_key(case_sensitive):
        return str if case_sensitive else str.casefold



from robocorp_ls_core.constants import *
from robocorp_ls_core.options import USE_TIMEOUTS
from typing import Optional

DEFAULT_COMPLETIONS_TIMEOUT: int = 4
if not USE_TIMEOUTS:
    # A whole month of timeout seems good enough as a max.
    DEFAULT_COMPLETIONS_TIMEOUT = 60 * 60 * 24 * 30



from robot.api.parsing import ModelTransformer, SectionHeader, Token

from robotidy.decorators import check_start_end_line


class NormalizeSectionHeaderName(ModelTransformer):
    """
    Normalize section headers names.
    Robot Framework is quite flexible with the section header naming. Following lines are equal:

        *setting
        *** SETTINGS
        *** SettingS ***

    This transformer normalizes naming to follow ``*** SectionName ***`` format (with plural variant):

        *** Settings ***
        *** Keywords ***
        *** Test Cases ***
        *** Variables ***
        *** Comments ***

    Optional data after section header (for example data driven column names) is preserved.
    It is possible to upper case section header names by passing ``uppercase=True`` parameter:

        *** SETTINGS ***

    Supports global formatting params: ``--startline`` and ``--endline``.

    See https://robotidy.readthedocs.io/en/latest/transformers/NormalizeSectionHeaderName.html for more examples.
    """

    def __init__(self, uppercase: bool = False):
        self.uppercase = uppercase

    @check_start_end_line
    def visit_SectionHeader(self, node):  # noqa
        if node.name and "task" in node.name.lower():
            name = "*** Tasks ***"
        else:
            name = {
                Token.SETTING_HEADER: "*** Settings ***",
                Token.VARIABLE_HEADER: "*** Variables ***",
                Token.TESTCASE_HEADER: "*** Test Cases ***",
                Token.KEYWORD_HEADER: "*** Keywords ***",
                Token.COMMENT_HEADER: "*** Comments ***",
            }[node.type]
        if self.uppercase:
            name = name.upper()
        # we only modify header token value in order to preserver optional data driven testing column names
        node.data_tokens[0].value = name
        return node



from __future__ import annotations

import os
import queue
import threading
import typing

from robocorp_ls_core.debug_adapter_core.dap.dap_base_schema import BaseSchema
from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
    TerminatedEvent,
    TerminatedEventBody,
    Event,
)
from robocorp_ls_core.robotframework_log import get_logger
from robotframework_debug_adapter.base_launch_process_target import (
    BaseLaunchProcessTargetComm,
)
from robotframework_ls.options import DEFAULT_TIMEOUT
from typing import Optional


if typing.TYPE_CHECKING:
    from robotframework_debug_adapter.debug_adapter_comm import DebugAdapterComm


log = get_logger(__name__)


class LaunchProcessDebugAdapterRobotTargetComm(BaseLaunchProcessTargetComm):
    """
    This class is used so intermediate talking to the server.

    It's the middle ground between the `DebugAdapterComm` and `_RobotTargetComm`.
        - `DebugAdapterComm`:
            It's used to talk with the client (in this process) and accessed
            through the _weak_debug_adapter_comm attribute.

        - `_RobotTargetComm`
            It's actually in the target process. We communicate with it by
            calling the `write_to_robot_message` method and receive messages
            from it in the `_from_robot` method in this class.
    """

    def __init__(self, debug_adapter_comm: DebugAdapterComm):
        BaseLaunchProcessTargetComm.__init__(self, debug_adapter_comm)

        self._server_socket = None

        self._write_to_robot_queue: "queue.Queue[BaseSchema]" = queue.Queue()

    def start_listening(self, connections_count):
        import socket

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind(("127.0.0.1", 0))
        port = s.getsockname()[1]
        # i.e. 2 connections accepted: robot target and pydevd (or only robot
        # if not in debug mode).
        s.listen(connections_count)
        self._server_socket = s
        self.start()
        return port, s

    def run(self):
        from robocorp_ls_core.debug_adapter_core.debug_adapter_threads import (
            writer_thread_no_auto_seq,
        )
        from robocorp_ls_core.debug_adapter_core.debug_adapter_threads import (
            reader_thread,
        )

        try:
            assert (
                self._server_socket is not None
            ), "start_listening must be called before start()"

            # while True:
            # Only handle a single connection...
            socket, _addr = self._server_socket.accept()

            read_from = socket.makefile("rb")
            write_to = socket.makefile("wb")

            debug_adapter_comm = self._weak_debug_adapter_comm()
            writer = self._writer_thread = threading.Thread(
                target=writer_thread_no_auto_seq,
                args=(write_to, self._write_to_robot_queue, "write to robot process"),
                name="Write to robot (LaunchProcessDebugAdapterRobotTargetComm)",
            )
            writer.daemon = True

            reader = self._reader_thread = threading.Thread(
                target=reader_thread,
                args=(
                    read_from,
                    self._from_robot,
                    debug_adapter_comm.write_to_client_queue,  # Used for errors
                    b"read from robot process",
                ),
                name="Read from robot (LaunchProcessDebugAdapterRobotTargetComm)",
            )
            reader.daemon = True

            reader.start()
            writer.start()

            self._connected_event.set()
        except:
            log.exception()

    def _from_robot(self, protocol_message: BaseSchema) -> None:
        self._handle_received_protocol_message_from_backend(protocol_message, "robot")

    def on_process_event(self, event):
        self._process_event_msg = event
        self._process_event.set()

        debug_adapter_comm = self._weak_debug_adapter_comm()
        if debug_adapter_comm is not None:
            event.body.kwargs["dapProcessId"] = os.getpid()
            debug_adapter_comm.write_to_client_message(event)
        else:
            log.debug("Command processor collected in event: %s" % (event,))

    def get_pid(self):
        assert self._process_event.is_set()
        return self._process_event_msg.body.systemProcessId

    def _forward_event_to_client(self, event: Event) -> None:
        debug_adapter_comm = self._weak_debug_adapter_comm()
        if debug_adapter_comm is not None:
            debug_adapter_comm.write_to_client_message(event)
        else:
            log.debug("Command processor collected in event: %s" % (event,))

    on_output_event = _forward_event_to_client

    on_startSuite_event = _forward_event_to_client
    on_endSuite_event = _forward_event_to_client

    on_startTest_event = _forward_event_to_client
    on_endTest_event = _forward_event_to_client

    on_logMessage_event = _forward_event_to_client

    def on_terminated_event(self, event: Optional[TerminatedEvent]) -> None:
        with self._terminated_lock:
            if self._terminated_event.is_set():
                return

            if event is None:
                restart = False
                event = TerminatedEvent(body=TerminatedEventBody(restart=restart))

            self._terminated_event_msg = event
            self._terminated_event.set()

        debug_adapter_comm = self._weak_debug_adapter_comm()
        if debug_adapter_comm is not None:
            debug_adapter_comm.write_to_client_message(event)
        else:
            log.debug("Command processor collected in event: %s" % (event,))

    def notify_exit(self) -> None:
        self.on_terminated_event(None)
        log.debug("Target process finished (forcibly exiting debug adapter in 100ms).")

        # If the target process is terminated, wait a bit and exit ourselves.
        import time

        time.sleep(0.1)
        os._exit(0)

    def write_to_robot_message(
        self, protocol_message: BaseSchema, on_response=None
    ) -> None:
        """
        :param BaseSchema protocol_message:
            Some instance of one of the messages in the debug_adapter.schema.
        """
        seq: int = self._next_seq()
        protocol_message.seq = seq
        if on_response is not None:
            self._msg_id_to_on_response[seq] = on_response
        self._write_to_robot_queue.put(protocol_message)

    def wait_for_connection(self):
        """
        :return bool:
            Returns True if the connection was successful and False otherwise.
        """
        assert self._server_socket is not None, "start_listening must be called first."
        log.debug("Wating for connection for %s seconds." % (DEFAULT_TIMEOUT,))
        ret = self._connected_event.wait(DEFAULT_TIMEOUT)
        log.debug("Connected: %s" % (ret,))
        return ret

    def wait_for_process_event(self):
        log.debug("Wating for process event for %s seconds." % (DEFAULT_TIMEOUT,))
        ret = self._process_event.wait(DEFAULT_TIMEOUT)
        log.debug("Received process event: %s" % (ret,))
        return ret



# coding: utf-8
from contextlib import contextmanager
import os

import pytest

from tests_python import debugger_unittest
from tests_python.debugger_unittest import (get_free_port, overrides, IS_CPYTHON, IS_JYTHON, IS_IRONPYTHON,
    CMD_ADD_DJANGO_EXCEPTION_BREAK, CMD_REMOVE_DJANGO_EXCEPTION_BREAK,
    CMD_ADD_EXCEPTION_BREAK, wait_for_condition, IS_PYPY)
from _pydevd_bundle.pydevd_comm_constants import file_system_encoding

import sys
from _pydevd_bundle.pydevd_constants import IS_WINDOWS


def get_java_location():
    from java.lang import System  # @UnresolvedImport
    jre_dir = System.getProperty("java.home")
    for f in [os.path.join(jre_dir, 'bin', 'java.exe'), os.path.join(jre_dir, 'bin', 'java')]:
        if os.path.exists(f):
            return f
    raise RuntimeError('Unable to find java executable')


def get_jython_jar():
    from java.lang import ClassLoader  # @UnresolvedImport
    cl = ClassLoader.getSystemClassLoader()
    paths = map(lambda url: url.getFile(), cl.getURLs())
    for p in paths:
        if 'jython.jar' in p:
            return p
    raise RuntimeError('Unable to find jython.jar')


class _WriterThreadCaseMSwitch(debugger_unittest.AbstractWriterThread):

    TEST_FILE = 'tests_python.resources._debugger_case_m_switch'
    IS_MODULE = True

    @overrides(debugger_unittest.AbstractWriterThread.get_environ)
    def get_environ(self):
        env = os.environ.copy()
        curr_pythonpath = env.get('PYTHONPATH', '')

        root_dirname = os.path.dirname(os.path.dirname(__file__))

        curr_pythonpath += root_dirname + os.pathsep
        env['PYTHONPATH'] = curr_pythonpath
        return env

    @overrides(debugger_unittest.AbstractWriterThread.get_main_filename)
    def get_main_filename(self):
        return debugger_unittest._get_debugger_test_file('_debugger_case_m_switch.py')


class _WriterThreadCaseModuleWithEntryPoint(_WriterThreadCaseMSwitch):

    TEST_FILE = 'tests_python.resources._debugger_case_module_entry_point:main'
    IS_MODULE = True

    @overrides(_WriterThreadCaseMSwitch.get_main_filename)
    def get_main_filename(self):
        return debugger_unittest._get_debugger_test_file('_debugger_case_module_entry_point.py')


class AbstractWriterThreadCaseFlask(debugger_unittest.AbstractWriterThread):

    FORCE_KILL_PROCESS_WHEN_FINISHED_OK = True
    FLASK_FOLDER = None

    TEST_FILE = 'flask'
    IS_MODULE = True

    def write_add_breakpoint_jinja2(self, line, func, template):
        '''
            @param line: starts at 1
        '''
        assert self.FLASK_FOLDER is not None
        breakpoint_id = self.next_breakpoint_id()
        template_file = debugger_unittest._get_debugger_test_file(os.path.join(self.FLASK_FOLDER, 'templates', template))
        self.write("111\t%s\t%s\t%s\t%s\t%s\t%s\tNone\tNone" % (self.next_seq(), breakpoint_id, 'jinja2-line', template_file, line, func))
        self.log.append('write_add_breakpoint_jinja: %s line: %s func: %s' % (breakpoint_id, line, func))
        return breakpoint_id

    def write_add_exception_breakpoint_jinja2(self, exception='jinja2-Exception'):
        self.write('%s\t%s\t%s\t%s\t%s\t%s' % (CMD_ADD_EXCEPTION_BREAK, self.next_seq(), exception, 2, 0, 0))

    @overrides(debugger_unittest.AbstractWriterThread.get_environ)
    def get_environ(self):
        import platform

        env = os.environ.copy()
        env['FLASK_APP'] = 'app.py'
        env['FLASK_ENV'] = 'development'
        env['FLASK_DEBUG'] = '0'
        if platform.system() != 'Windows':
            locale = 'en_US.utf8' if platform.system() == 'Linux' else 'en_US.UTF-8'
            env.update({
                'LC_ALL': locale,
                'LANG': locale,
            })
        return env

    def get_cwd(self):
        return debugger_unittest._get_debugger_test_file(self.FLASK_FOLDER)

    def get_command_line_args(self):
        assert self.FLASK_FOLDER is not None
        free_port = get_free_port()
        self.flask_port = free_port
        return [
            'flask',
            'run',
             '--no-debugger',
             '--no-reload',
             '--with-threads',
            '--port',
            str(free_port),
        ]

    def _ignore_stderr_line(self, line):
        if debugger_unittest.AbstractWriterThread._ignore_stderr_line(self, line):
            return True

        if 'Running on http:' in line:
            return True

        if 'GET / HTTP/' in line:
            return True

        return False

    def create_request_thread(self, url=''):
        return debugger_unittest.AbstractWriterThread.create_request_thread(
            self, 'http://127.0.0.1:%s%s' % (self.flask_port, url))


class AbstractWriterThreadCaseDjango(debugger_unittest.AbstractWriterThread):

    FORCE_KILL_PROCESS_WHEN_FINISHED_OK = True
    DJANGO_FOLDER = None

    def _ignore_stderr_line(self, line):
        if debugger_unittest.AbstractWriterThread._ignore_stderr_line(self, line):
            return True

        if 'GET /my_app' in line:
            return True

        return False

    def get_command_line_args(self):
        assert self.DJANGO_FOLDER is not None
        free_port = get_free_port()
        self.django_port = free_port
        return [
            debugger_unittest._get_debugger_test_file(os.path.join(self.DJANGO_FOLDER, 'manage.py')),
            'runserver',
            '--noreload',
            '--nothreading',
            str(free_port),
        ]

    def write_add_breakpoint_django(self, line, func, template):
        '''
            @param line: starts at 1
        '''
        assert self.DJANGO_FOLDER is not None
        breakpoint_id = self.next_breakpoint_id()
        template_file = debugger_unittest._get_debugger_test_file(os.path.join(self.DJANGO_FOLDER, 'my_app', 'templates', 'my_app', template))
        self.write("111\t%s\t%s\t%s\t%s\t%s\t%s\tNone\tNone" % (self.next_seq(), breakpoint_id, 'django-line', template_file, line, func))
        self.log.append('write_add_django_breakpoint: %s line: %s func: %s' % (breakpoint_id, line, func))
        return breakpoint_id

    def write_add_exception_breakpoint_django(self, exception='Exception'):
        self.write('%s\t%s\t%s' % (CMD_ADD_DJANGO_EXCEPTION_BREAK, self.next_seq(), exception))

    def write_remove_exception_breakpoint_django(self, exception='Exception'):
        self.write('%s\t%s\t%s' % (CMD_REMOVE_DJANGO_EXCEPTION_BREAK, self.next_seq(), exception))

    def create_request_thread(self, url=''):
        return debugger_unittest.AbstractWriterThread.create_request_thread(
            self, 'http://127.0.0.1:%s/%s' % (self.django_port, url))


class DebuggerRunnerSimple(debugger_unittest.DebuggerRunner):

    def get_command_line(self):
        if IS_JYTHON:
            if sys.executable is not None:
                # i.e.: we're running with the provided jython.exe
                return [sys.executable]
            else:

                return [
                    get_java_location(),
                    '-classpath',
                    get_jython_jar(),
                    'org.python.util.jython'
                ]

        if IS_CPYTHON or IS_PYPY:
            return [sys.executable, '-u']

        if IS_IRONPYTHON:
            return [
                    sys.executable,
                    '-X:Frames'
                ]

        raise RuntimeError('Unable to provide command line')


class DebuggerRunnerRemote(debugger_unittest.DebuggerRunner):

    def get_command_line(self):
        return [sys.executable, '-u']

    def add_command_line_args(self, args):
        writer = self.writer

        ret = args + [self.writer.TEST_FILE]
        ret = writer.update_command_line_args(ret)  # Provide a hook for the writer
        return ret


@pytest.fixture
def debugger_runner_simple(tmpdir):
    return DebuggerRunnerSimple(tmpdir)


@pytest.fixture
def debugger_runner_remote(tmpdir):
    return DebuggerRunnerRemote(tmpdir)


@pytest.fixture
def case_setup(tmpdir, debugger_runner_simple):
    runner = debugger_runner_simple

    class WriterThread(debugger_unittest.AbstractWriterThread):
        pass

    class CaseSetup(object):

        check_non_ascii = False
        NON_ASCII_CHARS = u'áéíóú汉字'

        @contextmanager
        def test_file(
                self,
                filename,
                wait_for_port=True,
                wait_for_initialization=True,
                **kwargs
            ):
            import shutil
            filename = debugger_unittest._get_debugger_test_file(filename)
            if self.check_non_ascii:
                basedir = str(tmpdir)
                if isinstance(basedir, bytes):
                    basedir = basedir.decode('utf-8')
                if isinstance(filename, bytes):
                    filename = filename.decode('utf-8')

                new_dir = os.path.join(basedir, self.NON_ASCII_CHARS)
                os.makedirs(new_dir)

                new_filename = os.path.join(new_dir, self.NON_ASCII_CHARS + os.path.basename(filename))
                shutil.copyfile(filename, new_filename)
                filename = new_filename

            WriterThread.TEST_FILE = filename
            for key, value in kwargs.items():
                assert hasattr(WriterThread, key)
                setattr(WriterThread, key, value)

            with runner.check_case(
                    WriterThread,
                    wait_for_port=wait_for_port,
                    wait_for_initialization=wait_for_initialization
                ) as writer:
                yield writer

    return CaseSetup()


@pytest.fixture
def case_setup_unhandled_exceptions(case_setup):

    original = case_setup.test_file

    def check_test_suceeded_msg(writer, stdout, stderr):
        return 'TEST SUCEEDED' in ''.join(stderr)

    def additional_output_checks(writer, stdout, stderr):
        # Don't call super as we have an expected exception
        if 'ValueError: TEST SUCEEDED' not in stderr:
            raise AssertionError('"ValueError: TEST SUCEEDED" not in stderr.\nstdout:\n%s\n\nstderr:\n%s' % (
                stdout, stderr))

    def test_file(*args, **kwargs):
        kwargs.setdefault('check_test_suceeded_msg', check_test_suceeded_msg)
        kwargs.setdefault('additional_output_checks', additional_output_checks)
        return original(*args, **kwargs)

    case_setup.test_file = test_file

    return case_setup


@pytest.fixture
def case_setup_remote(debugger_runner_remote):

    class WriterThread(debugger_unittest.AbstractWriterThread):
        pass

    class CaseSetup(object):

        @contextmanager
        def test_file(
                self,
                filename,
                wait_for_port=True,
                access_token=None,
                client_access_token=None,
                append_command_line_args=(),
                **kwargs
            ):

            def update_command_line_args(writer, args):
                ret = debugger_unittest.AbstractWriterThread.update_command_line_args(writer, args)
                wait_for_condition(lambda: hasattr(writer, 'port'))
                ret.append(str(writer.port))

                if access_token is not None:
                    ret.append('--access-token')
                    ret.append(access_token)
                if client_access_token is not None:
                    ret.append('--client-access-token')
                    ret.append(client_access_token)

                ret.extend(append_command_line_args)
                return ret

            WriterThread.TEST_FILE = debugger_unittest._get_debugger_test_file(filename)
            WriterThread.update_command_line_args = update_command_line_args
            for key, value in kwargs.items():
                assert hasattr(WriterThread, key)
                setattr(WriterThread, key, value)

            with debugger_runner_remote.check_case(WriterThread, wait_for_port=wait_for_port) as writer:
                yield writer

    return CaseSetup()


@pytest.fixture
def case_setup_remote_attach_to(debugger_runner_remote):
    '''
    The difference from this to case_setup_remote is that this one will connect to a server
    socket started by the debugger and case_setup_remote will create the server socket and wait
    for a connection from the debugger.
    '''

    class WriterThread(debugger_unittest.AbstractWriterThread):

        @overrides(debugger_unittest.AbstractWriterThread.run)
        def run(self):
            # I.e.: don't start socket on start(), rather, the test should call
            # start_socket_client() when needed.
            pass

    class CaseSetup(object):

        @contextmanager
        def test_file(
                self,
                filename,
                port,
                **kwargs
            ):
            additional_args = kwargs.pop('additional_args', [])

            def update_command_line_args(writer, args):
                ret = debugger_unittest.AbstractWriterThread.update_command_line_args(writer, args)
                ret.append(str(port))
                ret.extend(additional_args)
                return ret

            WriterThread.TEST_FILE = debugger_unittest._get_debugger_test_file(filename)
            WriterThread.update_command_line_args = update_command_line_args
            for key, value in kwargs.items():
                assert hasattr(WriterThread, key)
                setattr(WriterThread, key, value)

            with debugger_runner_remote.check_case(WriterThread, wait_for_port=False) as writer:
                yield writer

    return CaseSetup()


@pytest.fixture
def case_setup_multiprocessing(debugger_runner_simple):

    class WriterThread(debugger_unittest.AbstractWriterThread):
        pass

    class CaseSetup(object):

        @contextmanager
        def test_file(
                self,
                filename,
                **kwargs
            ):

            def update_command_line_args(writer, args):
                ret = debugger_unittest.AbstractWriterThread.update_command_line_args(writer, args)
                ret.insert(ret.index('--DEBUG_RECORD_SOCKET_READS'), '--multiprocess')
                return ret

            WriterThread.update_command_line_args = update_command_line_args
            WriterThread.TEST_FILE = debugger_unittest._get_debugger_test_file(filename)
            for key, value in kwargs.items():
                assert hasattr(WriterThread, key)
                setattr(WriterThread, key, value)

            with debugger_runner_simple.check_case(WriterThread) as writer:
                yield writer

    return CaseSetup()


@pytest.fixture
def case_setup_m_switch(debugger_runner_simple):

    class WriterThread(_WriterThreadCaseMSwitch):
        pass

    class CaseSetup(object):

        @contextmanager
        def test_file(self, **kwargs):
            for key, value in kwargs.items():
                assert hasattr(WriterThread, key)
                setattr(WriterThread, key, value)
            with debugger_runner_simple.check_case(WriterThread) as writer:
                yield writer

    return CaseSetup()


@pytest.fixture
def case_setup_m_switch_entry_point(debugger_runner_simple):

    runner = debugger_runner_simple

    class WriterThread(_WriterThreadCaseModuleWithEntryPoint):
        pass

    class CaseSetup(object):

        @contextmanager
        def test_file(self, **kwargs):
            for key, value in kwargs.items():
                assert hasattr(WriterThread, key)
                setattr(WriterThread, key, value)
            with runner.check_case(WriterThread) as writer:
                yield writer

    return CaseSetup()


@pytest.fixture
def case_setup_django(debugger_runner_simple):

    class WriterThread(AbstractWriterThreadCaseDjango):
        pass

    class CaseSetup(object):

        @contextmanager
        def test_file(self, **kwargs):
            import django
            version = [int(x) for x in django.get_version().split('.')][:2]
            if version == [1, 7]:
                django_folder = 'my_django_proj_17'
            elif version in ([2, 1], [2, 2], [3, 0], [3, 1], [3, 2], [4, 0]):
                django_folder = 'my_django_proj_21'
            else:
                raise AssertionError('Can only check django 1.7, 2.1, 2.2, 3.0, 3.1, 3.2 and 4.0 right now. Found: %s' % (version,))

            WriterThread.DJANGO_FOLDER = django_folder
            for key, value in kwargs.items():
                assert hasattr(WriterThread, key)
                setattr(WriterThread, key, value)

            with debugger_runner_simple.check_case(WriterThread) as writer:
                yield writer

    return CaseSetup()


@pytest.fixture
def case_setup_flask(debugger_runner_simple):

    class WriterThread(AbstractWriterThreadCaseFlask):
        pass

    class CaseSetup(object):

        @contextmanager
        def test_file(self, **kwargs):
            WriterThread.FLASK_FOLDER = 'flask1'
            for key, value in kwargs.items():
                assert hasattr(WriterThread, key)
                setattr(WriterThread, key, value)

            with debugger_runner_simple.check_case(WriterThread) as writer:
                yield writer

    return CaseSetup()



import pytest
import sys
from _pydevd_bundle.pydevd_constants import IS_JYTHON, IS_IRONPYTHON
from tests_python.debug_constants import TEST_CYTHON
from tests_python.debug_constants import PYDEVD_TEST_VM
import site
import os
from _pydev_bundle import pydev_log


def pytest_report_header(config):
    print('PYDEVD_USE_CYTHON: %s' % (TEST_CYTHON,))
    print('PYDEVD_TEST_VM: %s' % (PYDEVD_TEST_VM,))
    try:
        import multiprocessing
    except ImportError:
        pass
    else:
        print('Number of processors: %s' % (multiprocessing.cpu_count(),))

    print('Relevant system paths:')
    print('sys.executable: %s' % (sys.executable,))
    print('sys.prefix: %s' % (sys.prefix,))

    if hasattr(sys, 'base_prefix'):
        print('sys.base_prefix: %s' % (sys.base_prefix,))

    if hasattr(sys, 'real_prefix'):
        print('sys.real_prefix: %s' % (sys.real_prefix,))

    if hasattr(site, 'getusersitepackages'):
        print('site.getusersitepackages(): %s' % (site.getusersitepackages(),))

    if hasattr(site, 'getsitepackages'):
        print('site.getsitepackages(): %s' % (site.getsitepackages(),))

    for path in sys.path:
        if os.path.exists(path) and os.path.basename(path) == 'site-packages':
            print('Folder with "site-packages" in sys.path: %s' % (path,))


_started_monitoring_threads = False


def _start_monitoring_threads():
    # After the session finishes, wait 20 seconds to see if everything finished properly
    # and if it doesn't report an error.
    global _started_monitoring_threads
    if _started_monitoring_threads:
        return

    _started_monitoring_threads = True
    import threading
    if hasattr(sys, '_current_frames') and hasattr(threading, 'enumerate'):
        import time
        import traceback

        class DumpThreads(threading.Thread):

            def run(self):
                time.sleep(20)

                thread_id_to_name = {}
                try:
                    for t in threading.enumerate():
                        thread_id_to_name[t.ident] = '%s  (daemon: %s)' % (t.name, t.daemon)
                except:
                    pass

                stack_trace = [
                    '===============================================================================',
                    'pydev pyunit runner: Threads still found running after tests finished',
                    '================================= Thread Dump =================================']

                for thread_id, stack in sys._current_frames().items():
                    stack_trace.append('\n-------------------------------------------------------------------------------')
                    stack_trace.append(" Thread %s" % thread_id_to_name.get(thread_id, thread_id))
                    stack_trace.append('')

                    if 'self' in stack.f_locals:
                        sys.stderr.write(str(stack.f_locals['self']) + '\n')

                    for filename, lineno, name, line in traceback.extract_stack(stack):
                        stack_trace.append(' File "%s", line %d, in %s' % (filename, lineno, name))
                        if line:
                            stack_trace.append("   %s" % (line.strip()))
                stack_trace.append('\n=============================== END Thread Dump ===============================')
                sys.stderr.write('\n'.join(stack_trace))

                # Force thread run to finish
                import os
                os._exit(123)

        dump_current_frames_thread = DumpThreads()
        dump_current_frames_thread.daemon = True  # Daemon so that this thread doesn't halt it!
        dump_current_frames_thread.start()


def pytest_unconfigure():
    _start_monitoring_threads()


@pytest.fixture(scope="session", autouse=True)
def check_no_threads():
    yield
    _start_monitoring_threads()


# see: http://goo.gl/kTQMs
SYMBOLS = {
    'customary': ('B', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y'),
    'customary_ext': ('byte', 'kilo', 'mega', 'giga', 'tera', 'peta', 'exa',
                       'zetta', 'iotta'),
    'iec': ('Bi', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi', 'Yi'),
    'iec_ext': ('byte', 'kibi', 'mebi', 'gibi', 'tebi', 'pebi', 'exbi',
                       'zebi', 'yobi'),
}


def bytes2human(n, format='%(value).1f %(symbol)s', symbols='customary'):
    """
    Bytes-to-human / human-to-bytes converter.
    Based on: http://goo.gl/kTQMs
    Working with Python 2.x and 3.x.

    Author: Giampaolo Rodola' <g.rodola [AT] gmail [DOT] com>
    License: MIT
    """

    """
    Convert n bytes into a human readable string based on format.
    symbols can be either "customary", "customary_ext", "iec" or "iec_ext",
    see: http://goo.gl/kTQMs

      >>> bytes2human(0)
      '0.0 B'
      >>> bytes2human(0.9)
      '0.0 B'
      >>> bytes2human(1)
      '1.0 B'
      >>> bytes2human(1.9)
      '1.0 B'
      >>> bytes2human(1024)
      '1.0 K'
      >>> bytes2human(1048576)
      '1.0 M'
      >>> bytes2human(1099511627776127398123789121)
      '909.5 Y'

      >>> bytes2human(9856, symbols="customary")
      '9.6 K'
      >>> bytes2human(9856, symbols="customary_ext")
      '9.6 kilo'
      >>> bytes2human(9856, symbols="iec")
      '9.6 Ki'
      >>> bytes2human(9856, symbols="iec_ext")
      '9.6 kibi'

      >>> bytes2human(10000, "%(value).1f %(symbol)s/sec")
      '9.8 K/sec'

      >>> # precision can be adjusted by playing with %f operator
      >>> bytes2human(10000, format="%(value).5f %(symbol)s")
      '9.76562 K'
    """
    n = int(n)
    if n < 0:
        raise ValueError("n < 0")
    symbols = SYMBOLS[symbols]
    prefix = {}
    for i, s in enumerate(symbols[1:]):
        prefix[s] = 1 << (i + 1) * 10
    for symbol in reversed(symbols[1:]):
        if n >= prefix[symbol]:
            value = float(n) / prefix[symbol]
            return format % locals()
    return format % dict(symbol=symbols[0], value=n)


def format_memory_info(memory_info, curr_proc_memory_info):
    return 'Total: %s, Available: %s, Used: %s %%, Curr process: %s' % (
        bytes2human(memory_info.total), bytes2human(memory_info.available), memory_info.percent, format_process_memory_info(curr_proc_memory_info))


def format_process_memory_info(proc_memory_info):
    return bytes2human(proc_memory_info.rss)


DEBUG_MEMORY_INFO = False

_global_collect_info = False

PRINT_MEMORY_BEFORE_AFTER_TEST = False  # This makes running tests slower (but it may be handy to diagnose memory issues).


@pytest.fixture(autouse=PRINT_MEMORY_BEFORE_AFTER_TEST)
def before_after_each_function(request):
    global _global_collect_info

    try:
        import psutil  # Don't fail if not there
    except ImportError:
        yield
        return

    current_pids = set(proc.pid for proc in psutil.process_iter())
    before_curr_proc_memory_info = psutil.Process().memory_info()

    if _global_collect_info and DEBUG_MEMORY_INFO:
        try:
            from pympler import summary, muppy
            sum1 = summary.summarize(muppy.get_objects())
        except:
            pydev_log.exception()

    sys.stdout.write(
'''
===============================================================================
Memory before: %s
%s
===============================================================================
''' % (request.function, format_memory_info(psutil.virtual_memory(), before_curr_proc_memory_info)))
    yield

    processes_info = []
    for proc in psutil.process_iter():
        if proc.pid not in current_pids:
            try:
                try:
                    cmdline = proc.cmdline()
                except:
                    cmdline = '<unable to get>'
                processes_info.append(
                    'New Process: %s(%s - %s) - %s' % (
                        proc.name(),
                        proc.pid,
                        cmdline,
                        format_process_memory_info(proc.memory_info())
                    )
                )
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                pass  # The process could've died in the meanwhile

    after_curr_proc_memory_info = psutil.Process().memory_info()

    if DEBUG_MEMORY_INFO:
        try:
            if after_curr_proc_memory_info.rss - before_curr_proc_memory_info.rss > 10 * 1000 * 1000:
                # 10 MB leak
                if _global_collect_info:
                    sum2 = summary.summarize(muppy.get_objects())
                    diff = summary.get_diff(sum1, sum2)
                    sys.stdout.write('===============================================================================\n')
                    sys.stdout.write('Leak info:\n')
                    sys.stdout.write('===============================================================================\n')
                    summary.print_(diff)
                    sys.stdout.write('===============================================================================\n')

                _global_collect_info = True
                # We'll only really collect the info on the next test (i.e.: if at one test
                # we used too much memory, the next one will start collecting)
            else:
                _global_collect_info = False
        except:
            pydev_log.exception()

    sys.stdout.write(
'''
===============================================================================
Memory after: %s
%s%s
===============================================================================


''' % (
    request.function,
    format_memory_info(psutil.virtual_memory(), after_curr_proc_memory_info),
    '' if not processes_info else '\nLeaked processes:\n' + '\n'.join(processes_info)),
    )


from tests_python.regression_check import data_regression, datadir, original_datadir


@pytest.fixture
def pyfile(request, tmpdir):
    """
    Based on debugpy pyfile fixture (adapter for older versions of Python)

    A fixture providing a factory function that generates .py files.

    The returned factory takes a single function with an empty argument list,
    generates a temporary file that contains the code corresponding to the
    function body, and returns the full path to the generated file. Idiomatic
    use is as a decorator, e.g.:

        @pyfile
        def script_file():
            print('fizz')
            print('buzz')

    will produce a temporary file named script_file.py containing:

        print('fizz')
        print('buzz')

    and the variable script_file will contain the path to that file.

    In order for the factory to be able to extract the function body properly,
    function header ("def") must all be on a single line, with nothing after
    the colon but whitespace.

    Note that because the code is physically in a separate file when it runs,
    it cannot reuse top-level module imports - it must import all the modules
    that it uses locally. When linter complains, use #noqa.

    Returns a py.path.local instance that has the additional attribute "lines".
    After the source is writen to disk, tests.code.get_marked_line_numbers() is
    invoked on the resulting file to compute the value of that attribute.
    """
    import types
    import inspect

    def factory(source):
        assert isinstance(source, types.FunctionType)
        name = source.__name__
        source, _ = inspect.getsourcelines(source)

        # First, find the "def" line.
        def_lineno = 0
        for line in source:
            line = line.strip()
            if line.startswith("def") and line.endswith(":"):
                break
            def_lineno += 1
        else:
            raise ValueError("Failed to locate function header.")

        # Remove everything up to and including "def".
        source = source[def_lineno + 1:]
        assert source

        # Now we need to adjust indentation. Compute how much the first line of
        # the body is indented by, then dedent all lines by that amount. Blank
        # lines don't matter indentation-wise, and might not be indented to begin
        # with, so just replace them with a simple newline.
        for line in source:
            if line.strip():
                break  # i.e.: use first non-empty line
        indent = len(line) - len(line.lstrip())
        source = [l[indent:] if l.strip() else "\n" for l in source]
        source = "".join(source)

        # Write it to file.
        tmpfile = os.path.join(str(tmpdir), name + ".py")
        assert not os.path.exists(tmpfile), '%s already exists.' % (tmpfile,)
        with open(tmpfile, 'w') as stream:
            stream.write(source)

        return tmpfile

    return factory


if IS_JYTHON or IS_IRONPYTHON:

    # On Jython and IronPython, it's a no-op.
    def before_after_each_function():
        pass



from robot.api import SuiteVisitor
import os
import json
from typing import Set, Dict, Optional

from robocorp_ls_core.robotframework_log import get_logger
from robocorp_ls_core.basic import normalize_filename

log = get_logger(__name__)


class FilteringTestsSuiteVisitor(SuiteVisitor):
    def __init__(
        self, tests_filtering: Optional[Dict[str, Dict[str, Set[str]]]] = None
    ) -> None:
        super().__init__()
        # filename -> test names
        self.include: Dict[str, Set[str]] = {}
        self.exclude: Dict[str, Set[str]] = {}

        self._include_contains_cache: dict = {}
        self._exclude_contains_cache: dict = {}

        if tests_filtering is None:
            s = os.getenv("RFLS_PRERUN_FILTER_TESTS", "")
            if s:
                log.info("Found tests filtering: %s", s)
                tests_filtering = json.loads(s)

        def add(tup, container):
            source, test_name = tup
            source = self._normalize(source)
            s = container.get(source)
            if s is None:
                s = container[source] = set()
            s.add(test_name)

        if tests_filtering:
            for tup in tests_filtering.get("include", []):
                add(tup, self.include)
            for tup in tests_filtering.get("exclude", []):
                add(tup, self.exclude)

    def _normalize(self, source):
        return normalize_filename(source)

    def _contains(
        self, container: dict, source: str, test_name: str, cache: dict
    ) -> bool:
        # Note: we have a cache because _contains_uncached will always check
        # the parent structure for hits and whenever we find a hit we
        # can skip it.
        key = (source, test_name)
        ret = cache.get(key)
        if ret is not None:
            return ret

        ret = self._contains_uncached(container, source, test_name, cache)
        cache[key] = ret
        return ret

    def _contains_uncached(
        self, container: dict, source: str, test_name: str, cache: dict
    ) -> bool:
        # Try to check for the test directly
        test_names = container.get(source)
        if not test_names:
            dirname = os.path.dirname(source)
            if dirname == source or not dirname:
                return False

            return self._contains(
                container,
                dirname,
                "*",  # at a parent level the test name doesn't matter
                cache,
            )

        if "*" in test_names:
            return True
        if test_name != "*":
            return test_name in test_names
        return False

    def start_suite(self, suite) -> None:
        new_tests = []
        for t in suite.tests:
            source = self._normalize(t.source)

            if self.include:
                if not self._contains(
                    self.include, source, t.name, self._include_contains_cache
                ):
                    log.debug("Test not in includes: %s - %s", t.source, t.name)
                    continue

            # If we got here it's included, now, check excludes.
            if self.exclude:
                if self._contains(
                    self.exclude, source, t.name, self._exclude_contains_cache
                ):
                    log.debug("Test in excludes: %s - %s", t.source, t.name)
                    continue

            new_tests.append(t)
        suite.tests = new_tests

    def end_suite(self, suite):
        # We don't want to keep empty suites.
        suite.suites = [s for s in suite.suites if s.test_count > 0]



# Original work Copyright 2018 Palantir Technologies, Inc. (MIT)
# See ThirdPartyNotices.txt in the project root for license information.
# All modifications Copyright (c) Robocorp Technologies Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http: // www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import threading
from robocorp_ls_core.robotframework_log import get_logger
from typing import Optional
import json
from robocorp_ls_core.options import BaseOptions

log = get_logger(__name__)


def read(stream) -> Optional[str]:
    """
    Reads one message from the stream and returns the message (or None if EOF was reached).

    :param stream:
        The stream we should be reading from.

    :return str|NoneType:
        The message or None if the stream was closed.
    """
    headers = {}
    while True:
        # Interpret the http protocol headers
        line = stream.readline()  # The trailing \r\n should be there.

        if not line:  # EOF
            return None
        line = line.strip().decode("ascii")
        if not line:  # Read just a new line without any contents
            break
        try:
            name, value = line.split(": ", 1)
        except ValueError:
            raise RuntimeError("Invalid header line: {}.".format(line))
        headers[name.strip()] = value.strip()

    if not headers:
        raise RuntimeError("Got message without headers.")

    content_length = int(headers["Content-Length"])

    # Get the actual json
    body = _read_len(stream, content_length)
    return body.decode("utf-8")


def _read_len(stream, content_length) -> bytes:
    buf = b""
    if not content_length:
        return buf

    # Grab the body
    while True:
        data = stream.read(content_length - len(buf))
        if not buf and len(data) == content_length:
            # Common case
            return data
        buf += data
        if len(buf) == content_length:
            return buf
        if len(buf) > content_length:
            raise AssertionError(
                "Expected to read message up to len == %s (already read: %s). Found:\n%s"
                % (content_length, len(buf), buf.decode("utf-8", "replace"))
            )
        # len(buf) < content_length (just keep on going).


class JsonRpcStreamReader(object):
    def __init__(self, rfile):
        self._rfile = rfile

    def close(self):
        self._rfile.close()

    def listen(self, message_consumer):
        """Blocking call to listen for messages on the rfile.

        Args:
            message_consumer (fn): function that is passed each message as it is read off the socket.
        """
        try:
            while not self._rfile.closed:
                data = read(self._rfile)
                if data is None:
                    log.debug("Read: %s", data)
                    return

                try:
                    msg = json.loads(data)
                except:
                    log.exception("Failed to parse JSON message %s", data)
                    continue

                if isinstance(msg, dict):
                    if msg.get("command") not in BaseOptions.HIDE_COMMAND_MESSAGES:
                        log.debug("Read: %s", data)
                else:
                    log.debug("Read (non dict data): %s", data)

                try:
                    message_consumer(msg)
                except:
                    log.exception("Error processing JSON message %s", data)
                    continue
        except ConnectionResetError:
            pass  # Just ignore this one (connection was closed)
        except Exception:
            log.exception("Error in JsonRpcStreamReader.")
        finally:
            log.debug("Exited JsonRpcStreamReader.")


class JsonRpcStreamWriter(object):
    def __init__(self, wfile, **json_dumps_args):
        assert wfile is not None
        self._wfile = wfile
        self._wfile_lock = threading.Lock()
        self._json_dumps_args = json_dumps_args

    def close(self):
        log.debug("Will close writer")
        with self._wfile_lock:
            self._wfile.close()

    def write(self, message):
        with self._wfile_lock:
            if self._wfile.closed:
                log.debug("Unable to write %s (file already closed).", (message,))
                return False
            try:
                if isinstance(message, dict):
                    if message.get("command") not in BaseOptions.HIDE_COMMAND_MESSAGES:
                        log.debug("Writing: %s", message)
                else:
                    log.debug("Writing (non dict message): %s", message)

                body = json.dumps(message, **self._json_dumps_args)

                as_bytes = body.encode("utf-8")
                stream = self._wfile
                content_len_as_str = "Content-Length: %s\r\n\r\n" % len(as_bytes)
                content_len_bytes = content_len_as_str.encode("ascii")

                stream.write(content_len_bytes)
                stream.write(as_bytes)
                stream.flush()
                return True
            except Exception:  # pylint: disable=broad-except
                log.exception(
                    "Failed to write message to output file %s - closed: %s",
                    message,
                    self._wfile.closed,
                )
                return False



#!/usr/bin/env python
"""
A Simple wx example to test PyDev's event loop integration.

To run this:
1) Enable the PyDev GUI event loop integration for wx
2) do an execfile on this script
3) ensure you have a working GUI simultaneously with an
   interactive console

Ref: Modified from wxPython source code wxPython/samples/simple/simple.py
"""

if __name__ == '__main__':

    import wx
    
    
    class MyFrame(wx.Frame):
        """
        This is MyFrame.  It just shows a few controls on a wxPanel,
        and has a simple menu.
        """
        def __init__(self, parent, title):
            wx.Frame.__init__(self, parent, -1, title,
                              pos=(150, 150), size=(350, 200))
    
            # Create the menubar
            menuBar = wx.MenuBar()
    
            # and a menu
            menu = wx.Menu()
    
            # add an item to the menu, using \tKeyName automatically
            # creates an accelerator, the third param is some help text
            # that will show up in the statusbar
            menu.Append(wx.ID_EXIT, "E&xit\tAlt-X", "Exit this simple sample")
    
            # bind the menu event to an event handler
            self.Bind(wx.EVT_MENU, self.on_time_to_close, id=wx.ID_EXIT)
    
            # and put the menu on the menubar
            menuBar.Append(menu, "&File")
            self.SetMenuBar(menuBar)
    
            self.CreateStatusBar()
    
            # Now create the Panel to put the other controls on.
            panel = wx.Panel(self)
    
            # and a few controls
            text = wx.StaticText(panel, -1, "Hello World!")
            text.SetFont(wx.Font(14, wx.SWISS, wx.NORMAL, wx.BOLD))
            text.SetSize(text.GetBestSize())
            btn = wx.Button(panel, -1, "Close")
            funbtn = wx.Button(panel, -1, "Just for fun...")
    
            # bind the button events to handlers
            self.Bind(wx.EVT_BUTTON, self.on_time_to_close, btn)
            self.Bind(wx.EVT_BUTTON, self.on_fun_button, funbtn)
    
            # Use a sizer to layout the controls, stacked vertically and with
            # a 10 pixel border around each
            sizer = wx.BoxSizer(wx.VERTICAL)
            sizer.Add(text, 0, wx.ALL, 10)
            sizer.Add(btn, 0, wx.ALL, 10)
            sizer.Add(funbtn, 0, wx.ALL, 10)
            panel.SetSizer(sizer)
            panel.Layout()
    
    
        def on_time_to_close(self, evt):
            """Event handler for the button click."""
            print("See ya later!")
            self.Close()
    
        def on_fun_button(self, evt):
            """Event handler for the button click."""
            print("Having fun yet?")
    
    
    class MyApp(wx.App):
        def OnInit(self):
            frame = MyFrame(None, "Simple wxPython App")
            self.SetTopWindow(frame)
    
            print("Print statements go to this stdout window by default.")
    
            frame.Show(True)
            return True
    
    
    if __name__ == '__main__':
    
        app = wx.GetApp()
        if app is None:
            app = MyApp(redirect=False, clearSigInt=False)
        else:
            frame = MyFrame(None, "Simple wxPython App")
            app.SetTopWindow(frame)
            print("Print statements go to this stdout window by default.")
            frame.Show(True)
    



def method1(n):
    if n <= 0:
        return 0  # Break here
    method2(n - 1)


def method2(n):
    method1(n - 1)


if __name__ == '__main__':
    try:
        method1(100)
    except:
        pass  # Don't let it print the exception (just deal with caught exceptions).
    print('TEST SUCEEDED!')



# Original work Copyright (c) 2012, C. G. Wrench <c.g.wrench@gmail.com>
# All modifications Copyright (c) Robocorp Technologies Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http: // www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# -*- coding: utf-8 -*-
"""Simple Markdown writer for reStructuredText.

See:

* https://daringfireball.net/projects/markdown/syntax
* https://guides.github.com/pdfs/markdown-cheatsheet-online.pdf
* https://github.com/matthew-brett/nb2plots/blob/master/nb2plots/doctree2md.py
"""
from __future__ import unicode_literals

import re
from textwrap import dedent
import posixpath

__docformat__ = "reStructuredText"

from docutils import frontend, nodes, writers, languages
from collections import OrderedDict


class IndentLevel(object):
    """Class to hold text being written for a certain indentation level

    For example, all text in list_elements need to be indented.  A list_element
    creates one of these indentation levels, and all text contained in the
    list_element gets written to this IndentLevel.  When we leave the
    list_element, we ``write`` the text with suitable prefixes to the next
    level down, which might be the base of the document (document body) or
    another indentation level, if this is - for example - a nested list.

    In most respects, IndentLevel behaves like a list.
    """

    def __init__(self, base, prefix, first_prefix=None):
        self.base = base  # The list to which we eventually write
        self.prefix = prefix  # Text prepended to lines
        # Text prepended to first list
        self.first_prefix = prefix if first_prefix is None else first_prefix
        # Our own list to which we append before doing a ``write``
        self.content = []

    def append(self, new):
        self.content.append(new)

    def __getitem__(self, index):
        return self.content[index]

    def __len__(self):
        return len(self.content)

    def __bool__(self):
        return len(self) != 0

    def write(self):
        """Add ``self.contents`` with current ``prefix`` and ``first_prefix``

        Add processed ``self.contents`` to ``self.base``.  The first line has
        ``first_prefix`` prepended, further lines have ``prefix`` prepended.

        Empty (all whitespace) lines get written as bare carriage returns, to
        avoid ugly extra whitespace.
        """
        string = "".join(self.content)
        lines = string.splitlines(True)
        if len(lines) == 0:
            return
        texts = [self.first_prefix + lines[0]]
        for line in lines[1:]:
            if line.strip() == "":  # avoid prefix for empty lines
                texts.append("\n")
            else:
                texts.append(self.prefix + line)
        self.base.append("".join(texts))


def _make_method(to_add):
    """Make a method that adds `to_add`

    We need this function so that `to_add` is a fresh and unique variable at
    the time the method is defined.
    """

    def method(self, node):
        self.add(to_add)

    return method


def add_pref_suff(pref_suff_map):
    """Decorator adds visit, depart methods for prefix/suffix pairs"""

    def dec(cls):
        # Need _make_method to ensure new variable picked up for each iteration
        # of the loop.  The defined method picks up this new variable in its
        # scope.
        for key, (prefix, suffix) in pref_suff_map.items():
            setattr(cls, "visit_" + key, _make_method(prefix))
            setattr(cls, "depart_" + key, _make_method(suffix))
        return cls

    return dec


def add_pass_thru(pass_thrus):
    """Decorator adds explicit pass-through visit and depart methods"""

    def meth(self, node):
        pass

    def dec(cls):
        for element_name in pass_thrus:
            for meth_prefix in ("visit_", "depart_"):
                meth_name = meth_prefix + element_name
                if hasattr(cls, meth_name):
                    raise ValueError("method name {} already defined".format(meth_name))
                setattr(cls, meth_name, meth)
        return cls

    return dec


# Characters that should be escaped in Markdown
ESCAPE_RE = re.compile(r"([\\*`])")

# Doctree elements for which Markdown element is <prefix><content><suffix>
PREF_SUFF_ELEMENTS = {
    "emphasis": ("*", "*"),  # Could also use ('_', '_')
    "strong": ("**", "**"),  # Could also use ('__', '__')
    "literal": ("`", "`"),
    "subscript": ("<sub>", "</sub>"),
    "superscript": ("<sup>", "</sup>"),
}

# Doctree elements explicitly passed through without extra markup
PASS_THRU_ELEMENTS = (
    "document",
    "container",
    "target",
    "inline",
    "definition_list",
    "definition_list_item",
    "term",
    "field_list",
    "field_list_item",
    "field",
    "field_name",
    "mpl_hint",
    "pending_xref",
    "compound",
)


@add_pass_thru(PASS_THRU_ELEMENTS)
@add_pref_suff(PREF_SUFF_ELEMENTS)
class Translator(nodes.NodeVisitor):

    std_indent = "    "

    def __init__(self, document, builder=None):
        nodes.NodeVisitor.__init__(self, document)
        self.builder = builder
        self.settings = settings = document.settings
        lcode = settings.language_code
        self.language = languages.get_language(lcode, document.reporter)
        # Not-None here indicates Markdown should use HTTP for internal and
        # download links.
        self.markdown_http_base = builder.markdown_http_base if builder else None
        # Warn only once per writer about unsupported elements
        self._warned = set()
        # Lookup table to get section list from name
        self._lists = OrderedDict((("head", []), ("body", []), ("foot", [])))
        # Reset attributes modified by reading
        self.reset()
        # Attribute shortcuts
        self.head, self.body, self.foot = self._lists.values()

    def reset(self):
        """Initialize object for fresh read"""
        for part in self._lists.values():
            part[:] = []

        # Current section heading level during writing
        self.section_level = 0

        # FIFO list of list prefixes, while writing nested lists.  Each element
        # corresponds to one level of nesting.  Thus ['1. ', '1. ', '* '] would
        # occur when writing items of an unordered list, that is nested within
        # an ordered list, that in turn is nested in another ordered list.
        self.list_prefixes = []

        # FIFO list of indentation levels.  When we are writing a block of text
        # that should be indented, we create a new indentation level.  We only
        # write the text when we leave the indentation level, so we can insert
        # the correct prefix for every line.
        self.indent_levels = []

        # Flag indicating we are processing docinfo items
        self._in_docinfo = False

        # Flag for whether to escape characters
        self._escape_text = True

    def astext(self):
        """Return the final formatted document as a string."""
        parts = ["".join(lines).strip() for lines in self._lists.values()]
        parts = [part + "\n\n" for part in parts if part]
        return "".join(parts).strip() + "\n"

    def ensure_eol(self):
        """Ensure the last line in current base is terminated by new line."""
        out = self.get_current_output()
        if out and out[-1] and out[-1][-1] != "\n":
            out.append("\n")

    def get_current_output(self, section="body"):
        """Get list or IndentLevel to which we are currently writing"""
        return self.indent_levels[-1] if self.indent_levels else self._lists[section]

    def add(self, string, section="body"):
        """Add `string` to `section` or current output

        Parameters
        ----------
        string : str
            String to add to output document
        section : {'body', 'head', 'foot'}, optional
            Section of document that generated text should be appended to, if
            not already appending to an indent level.  If already appending to
            an indent level, method will ignore `section`.  Use
            :meth:`add_section` to append to a section unconditionally.
        """
        self.get_current_output(section).append(string)

    def add_section(self, string, section="body"):
        """Add `string` to `section` regardless of current output

        Can be useful when forcing write to header or footer.

        Parameters
        ----------
        string : str
            String to add to output document
        section : {'body', 'head', 'foot'}, optional
            Section of document that generated text should be appended to.
        """
        self._lists[section].append(string)

    def start_level(self, prefix, first_prefix=None, section="body"):
        """Create a new IndentLevel with `prefix` and `first_prefix`"""
        base = (
            self.indent_levels[-1].content
            if self.indent_levels
            else self._lists[section]
        )
        level = IndentLevel(base, prefix, first_prefix)
        self.indent_levels.append(level)

    def finish_level(self):
        """Remove most recent IndentLevel and write contents"""
        level = self.indent_levels.pop()
        level.write()

    def escape_chars(self, txt):
        # Escape (some) characters with special meaning for Markdown
        return ESCAPE_RE.sub(r"\\\1", txt)

    def visit_Text(self, node):
        text = node.astext()
        if self._escape_text:
            text = self.escape_chars(text)
        self.add(text)

    def depart_Text(self, node):
        pass

    def visit_comment(self, node):
        txt = node.astext()
        if txt.strip():
            self.add("<!-- " + node.astext() + " -->\n")
        raise nodes.SkipNode

    def visit_docinfo(self, node):
        self._in_docinfo = True

    def depart_docinfo(self, node):
        self._in_docinfo = False

    def process_docinfo_item(self, node):
        """Called explicitly from methods in this class"""
        self.add_section("% {}\n".format(node.astext()), section="head")
        raise nodes.SkipNode

    def visit_definition(self, node):
        self.add("\n\n")
        self.start_level("    ", ":   ")

    def depart_definition(self, node):
        self.finish_level()

    visit_field_body = visit_definition

    depart_field_body = depart_definition

    def visit_paragraph(self, node):
        pass

    def depart_paragraph(self, node):
        self.ensure_eol()
        self.add("\n")

    def visit_math_block(self, node):
        # docutils math block
        self._escape_text = False
        self.add("$$\n")

    def depart_math_block(self, node):
        self._escape_text = True
        self.ensure_eol()
        self.add("$$\n\n")

    def visit_displaymath(self, node):
        # sphinx math blocks become displaymath
        self.add("$$\n{}\n$$\n\n".format(node["latex"]))
        raise nodes.SkipNode

    def visit_math(self, node):
        # sphinx math node has 'latex' attribute, docutils does not
        if "latex" in node:  # sphinx math node
            self.add("${}$".format(node["latex"]))
            raise nodes.SkipNode
        # docutils math node
        self._escape_text = False
        self.add("$")

    def depart_math(self, node):
        # sphinx node skipped in visit, only docutils gets here
        self._escape_text = True
        self.add("$")

    def visit_literal_block(self, node):
        self._escape_text = False
        if "code" in node["classes"]:  # Sphinx < 2
            if len(node["classes"]) > 1:
                code_type = node["classes"][1]
            else:
                code_type = ""
        else:  # Sphinx >= 2
            language = node.get("language", "")
            code_type = "" if language == "default" else language
        self.add("```" + code_type + "\n")

    def depart_literal_block(self, node):
        self._escape_text = True
        self.ensure_eol()
        self.add("```\n\n")

    def visit_doctest_block(self, node):
        self._escape_text = False
        self.add("```python\n")

    depart_doctest_block = depart_literal_block

    def visit_block_quote(self, node):
        self.start_level("> ")

    def depart_block_quote(self, node):
        self.finish_level()

    def visit_section(self, node):
        self.section_level += 1

    def depart_section(self, node):
        self.section_level -= 1

    def visit_enumerated_list(self, node):
        self.list_prefixes.append("1. ")

    def depart_enumerated_list(self, node):
        self.list_prefixes.pop()

    def visit_bullet_list(self, node):
        self.list_prefixes.append("* ")

    depart_bullet_list = depart_enumerated_list

    def visit_list_item(self, node):
        first_prefix = self.list_prefixes[-1]
        prefix = " " * len(first_prefix)
        self.start_level(prefix, first_prefix)

    def depart_list_item(self, node):
        self.finish_level()

    def visit_problematic(self, node):
        self.add("\n\n```\n{}\n```\n\n".format(node.astext()))
        raise nodes.SkipNode

    def visit_system_message(self, node):
        if node["level"] < self.document.reporter.report_level:
            # Level is too low to display
            raise nodes.SkipNode
        line = ", line %s" % node["line"] if node.hasattr("line") else ""
        self.add(
            "```\nSystem Message: {}:{}\n\n{}\n```\n\n".format(
                node["source"], line, node.astext()
            )
        )
        raise nodes.SkipNode

    def visit_title(self, node):
        self.add((self.section_level + 1) * "#" + " ")

    def depart_title(self, node):
        self.ensure_eol()
        self.add("\n")

    def visit_subtitle(self, node):
        self.add((self.section_level + 2) * "#" + " ")

    depart_subtitle = depart_title

    def visit_transition(self, node):
        # Simply replace a transition by a horizontal rule.
        # Could use three or more '*', '_' or '-'.
        self.add("\n---\n\n")
        raise nodes.SkipNode

    def _refuri2http(self, node):
        # Replace 'refuri' in reference with HTTP address, if possible
        # None for no possible address
        url = node.get("refuri")
        if not node.get("internal"):
            return url
        # If HTTP page build URL known, make link relative to that.
        if not self.markdown_http_base:
            return None
        this_doc = self.builder.current_docname
        if url in (None, ""):  # Reference to this doc
            url = self.builder.get_target_uri(this_doc)
        else:  # URL is relative to the current docname.
            this_dir = posixpath.dirname(this_doc)
            if this_dir:
                url = posixpath.normpath("{}/{}".format(this_dir, url))
        url = "{}/{}".format(self.markdown_http_base, url)
        if "refid" in node:
            url += "#" + node["refid"]
        return url

    def visit_reference(self, node):
        # If no target possible, pass through.
        url = self._refuri2http(node)
        if url is None:
            return
        self.add("[{0}]({1})".format(node.astext(), url))
        raise nodes.SkipNode

    def depart_reference(self, node):
        pass

    def visit_code_links(self, node):
        raise nodes.SkipNode

    def visit_index(self, node):
        # Drop index entries
        raise nodes.SkipNode

    def visit_substitution_definition(self, node):
        # Drop substitution definitions - the doctree already contains the text
        # with substitutions applied.
        raise nodes.SkipNode

    def visit_only(self, node):
        if node["expr"] == "markdown":
            self.add(dedent(node.astext()) + "\n")
        raise nodes.SkipNode

    def visit_runrole_reference(self, node):
        raise nodes.SkipNode

    def visit_download_reference(self, node):
        # If not resolving internal links, or there is no filename specified,
        # pass through.
        filename = node.get("filename")
        if None in (self.markdown_http_base, filename):
            return
        target_url = "{}/_downloads/{}".format(self.markdown_http_base, filename)
        self.add("[{}]({})".format(node.astext(), target_url))
        raise nodes.SkipNode

    def depart_download_reference(self, node):
        pass

    visit_compact_paragraph = visit_paragraph

    depart_compact_paragraph = depart_paragraph

    def unknown_visit(self, node):
        """Warn once per instance for unsupported nodes

        Intercept docinfo items if in docinfo block.
        """
        if self._in_docinfo:
            self.process_docinfo_item(node)
            return

        # Add as plain text on unknown.
        self.add(_escape_markdown(node.astext()))
        raise nodes.SkipNode

        # We really don't know this node type, warn once per node type
        # node_type = node.__class__.__name__
        # if node_type not in self._warned:
        #     self.document.reporter.warning(
        #         "The " + node_type + " element not yet supported in Markdown."
        #     )
        #     self._warned.add(node_type)


def _escape_markdown(text):
    if not text:
        return ""
    return (
        text.replace("_", r"\_")
        .replace("*", r"\*")
        .replace("[", r"\[")
        .replace("]", r"\]")
        .replace("#", r"\#")
    )


class Writer(writers.Writer):

    supported = ("markdown",)
    """Formats this writer supports."""

    output = None
    """Final translated form of `document`."""

    # Add configuration settings for additional Markdown flavours here.
    settings_spec = (
        "Markdown-Specific Options",
        None,
        (
            (
                "Extended Markdown syntax.",
                ["--extended-markdown"],
                {
                    "default": 0,
                    "action": "store_true",
                    "validator": frontend.validate_boolean,
                },
            ),
            (
                "Strict Markdown syntax. Default: true",
                ["--strict-markdown"],
                {
                    "default": 1,
                    "action": "store_true",
                    "validator": frontend.validate_boolean,
                },
            ),
        ),
    )

    translator_class = Translator

    def __init__(self, builder=None):
        writers.Writer.__init__(self)
        self.builder = builder

    def translate(self):
        visitor = self.translator_class(self.document, self.builder)
        self.document.walkabout(visitor)
        self.output = visitor.astext()



from robocorp_ls_core.python_ls import PythonLanguageServer
from robocorp_ls_core.basic import overrides, log_and_silence_errors
import os
import time
from robotframework_ls.constants import DEFAULT_COMPLETIONS_TIMEOUT
from robocorp_ls_core.robotframework_log import get_logger
from typing import Any, Optional, Dict
from robocorp_ls_core.protocols import (
    IConfig,
    IWorkspace,
    IIdMessageMatcher,
    IRobotFrameworkApiClient,
    IMonitor,
)
from pathlib import Path
from robotframework_ls.ep_providers import (
    EPConfigurationProvider,
    EPDirCacheProvider,
    EPEndPointProvider,
)
from robocorp_ls_core.jsonrpc.endpoint import require_monitor
from functools import partial
import itertools
from robotframework_ls import __version__, rf_interactive_integration
import typing
import sys
from robocorp_ls_core.watchdog_wrapper import IFSObserver
from robocorp_ls_core.lsp import (
    CodeLensTypedDict,
    TextDocumentPositionParamsTypedDict,
    PositionTypedDict,
    CompletionItemTypedDict,
)
from robotframework_ls.commands import (
    ROBOT_GET_RFLS_HOME_DIR,
    ROBOT_START_INDEXING_INTERNAL,
    ROBOT_WAIT_FULL_TEST_COLLECTION_INTERNAL,
)


log = get_logger(__name__)

_LINT_DEBOUNCE_IN_SECONDS_LOW = 0.2
_LINT_DEBOUNCE_IN_SECONDS_HIGH = 0.8


class _CurrLintInfo(object):
    def __init__(
        self,
        rf_lint_api_client: IRobotFrameworkApiClient,
        lsp_messages,
        doc_uri,
        is_saved,
        on_finish,
    ) -> None:
        from robocorp_ls_core.lsp import LSPMessages
        from robocorp_ls_core.jsonrpc.monitor import Monitor

        self._rf_lint_api_client = rf_lint_api_client
        self.lsp_messages: LSPMessages = lsp_messages
        self.doc_uri = doc_uri
        self.is_saved = is_saved
        self._monitor = Monitor()
        self._on_finish = on_finish

    def __call__(self) -> None:
        from robocorp_ls_core.jsonrpc.exceptions import JsonRpcRequestCancelled
        from robocorp_ls_core.client_base import wait_for_message_matcher
        from robotframework_ls.server_api.client import SubprocessDiedError

        try:
            doc_uri = self.doc_uri
            self._monitor.check_cancelled()
            found = []
            message_matcher = self._rf_lint_api_client.request_lint(doc_uri)
            if message_matcher is not None:
                if wait_for_message_matcher(
                    message_matcher,
                    monitor=self._monitor,
                    request_cancel=self._rf_lint_api_client.request_cancel,
                    timeout=60 * 3,
                ):
                    diagnostics_msg = message_matcher.msg
                    if diagnostics_msg:
                        found = diagnostics_msg.get("result", [])
                    self._monitor.check_cancelled()
                    self.lsp_messages.publish_diagnostics(doc_uri, found)
        except JsonRpcRequestCancelled:
            log.debug("Cancelled linting: %s.", self.doc_uri)

        except SubprocessDiedError:
            log.debug("Subprocess exited while linting: %s.", self.doc_uri)

        except Exception:
            log.exception("Error linting: %s.", self.doc_uri)

        finally:
            self._on_finish(self)

    def cancel(self):
        self._monitor.cancel()


def run_in_new_thread(func, thread_name):
    import threading

    t = threading.Thread(target=func)
    t.name = thread_name
    t.start()


class _LintManager(object):
    def __init__(self, server_manager, lsp_messages) -> None:
        import threading
        from robotframework_ls.server_manager import ServerManager

        self._server_manager: ServerManager = server_manager
        self._lsp_messages = lsp_messages

        self._next_id = partial(next, itertools.count())
        self._doc_id_to_info: Dict[str, _CurrLintInfo] = {}
        self._lock = threading.Lock()

    def schedule_lint(self, doc_uri: str, is_saved: bool, timeout: float) -> None:
        self.cancel_lint(doc_uri)
        rf_lint_api_client = self._server_manager.get_lint_rf_api_client(doc_uri)
        if rf_lint_api_client is None:
            log.info("Unable to get lint api for: %s", doc_uri)
            return

        lock = self._lock
        doc_id_to_info = self._doc_id_to_info

        def on_finish(curr_info: _CurrLintInfo):
            # When it's finished, remove it from our references (if it's still
            # the current one...).
            try:
                with lock:
                    if doc_id_to_info.get(curr_info.doc_uri) is curr_info:
                        del doc_id_to_info[curr_info.doc_uri]
            except:
                log.exception("Unhandled error on lint finish.")

        log.debug("Schedule lint for: %s", doc_uri)
        curr_info = _CurrLintInfo(
            rf_lint_api_client, self._lsp_messages, doc_uri, is_saved, on_finish
        )
        with self._lock:
            self._doc_id_to_info[doc_uri] = curr_info

        from robocorp_ls_core.timeouts import TimeoutTracker

        timeout_tracker = TimeoutTracker.get_singleton()
        timeout_tracker.call_on_timeout(
            timeout,
            partial(run_in_new_thread, curr_info, f"Lint: {doc_uri}"),
        )

    def cancel_lint(self, doc_uri: str) -> None:
        with self._lock:
            curr_info = self._doc_id_to_info.pop(doc_uri, None)
            if curr_info is not None:
                log.debug("Cancel lint for: %s", doc_uri)

                curr_info.cancel()


from robocorp_ls_core.command_dispatcher import _CommandDispatcher

command_dispatcher = _CommandDispatcher()


class RobotFrameworkLanguageServer(PythonLanguageServer):
    def __init__(self, rx, tx) -> None:
        from robocorp_ls_core.pluginmanager import PluginManager
        from robotframework_ls.rf_interactive_integration import _RfInterpretersManager
        from robotframework_ls.server_manager import ServerManager
        from robotframework_ls.ep_providers import DefaultConfigurationProvider
        from robotframework_ls.ep_providers import DefaultEndPointProvider
        from robotframework_ls.ep_providers import DefaultDirCacheProvider
        from robocorp_ls_core import watchdog_wrapper
        from robocorp_ls_core.remote_fs_observer_impl import RemoteFSObserver
        from robocorp_ls_core.options import Setup
        from robotframework_ls.robotframework_ls_completion_impl import (
            _RobotFrameworkLsCompletionImpl,
        )

        PythonLanguageServer.__init__(self, rx, tx)

        from robocorp_ls_core.cache import DirCache

        from robotframework_ls import robot_config

        home = robot_config.get_robotframework_ls_home()
        cache_dir = os.path.join(home, ".cache")

        log.debug(f"Cache dir: {cache_dir}")

        self._last_doc_uri: str = ""
        self._dir_cache = DirCache(cache_dir)

        self._pm = PluginManager()
        self._config_provider = DefaultConfigurationProvider(self.config)
        self._pm.set_instance(EPConfigurationProvider, self._config_provider)
        self._pm.set_instance(
            EPDirCacheProvider, DefaultDirCacheProvider(self._dir_cache)
        )
        self._pm.set_instance(
            EPEndPointProvider, DefaultEndPointProvider(self._endpoint)
        )
        self._rf_interpreters_manager = _RfInterpretersManager(self._endpoint, self._pm)

        watch_impl = os.environ.get("ROBOTFRAMEWORK_LS_WATCH_IMPL", "auto")
        if watch_impl not in ("watchdog", "fsnotify", "auto"):
            log.info(
                f"ROBOTFRAMEWORK_LS_WATCH_IMPL should be 'auto', 'watchdog' or 'fsnotify'. Found: {watch_impl} (falling back to auto)"
            )
            watch_impl = "auto"

        if watch_impl == "auto":
            # In auto mode we use watchdog for windows and fsnotify (polling)
            # for Linux and Mac. The reason for that is that on Linux and Mac
            # if big folders are watched the system may complain due to the
            # lack of resources, which may prevent the extension from working
            # properly.
            #
            # If users want to opt-in, they can change to watchdog (and
            # ideally install it to their env to get native extensions).
            if sys.platform == "win32":
                watch_impl = "watchdog"
            else:
                watch_impl = "fsnotify"

        log.info(
            "Using watch implementation: %s (customize with ROBOTFRAMEWORK_LS_WATCH_IMPL environment variable)",
            watch_impl,
        )

        self._fs_observer = watchdog_wrapper.create_remote_observer(
            watch_impl, (".py", ".libspec", "robot", ".resource")
        )
        remote_observer = typing.cast(RemoteFSObserver, self._fs_observer)

        log_file = Setup.options.log_file
        verbose = Setup.options.verbose
        if not isinstance(log_file, str) and log_file is not None:
            log.critical("Expected %r to be a str (was: %s)", log_file, type(log_file))
            log_file = None

        if not isinstance(verbose, int) and verbose is not None:
            log.critical("Expected %r to be an int (was: %s)", verbose, type(verbose))
            verbose = None

        remote_observer.start_server(log_file=log_file, verbose=verbose)

        self._server_manager = ServerManager(self._pm, language_server=self)
        self._lint_manager = _LintManager(self._server_manager, self._lsp_messages)
        self._robot_framework_ls_completion_impl = _RobotFrameworkLsCompletionImpl(
            self._server_manager, self
        )

    def get_remote_fs_observer_port(self) -> Optional[int]:
        from robocorp_ls_core.remote_fs_observer_impl import RemoteFSObserver

        remote_observer = typing.cast(RemoteFSObserver, self._fs_observer)
        return remote_observer.port

    @overrides(PythonLanguageServer._create_config)
    def _create_config(self) -> IConfig:
        from robotframework_ls.robot_config import RobotConfig

        return RobotConfig()

    @overrides(PythonLanguageServer._on_workspace_set)
    def _on_workspace_set(self, workspace: IWorkspace):
        PythonLanguageServer._on_workspace_set(self, workspace)
        self._server_manager.set_workspace(workspace)

    @overrides(PythonLanguageServer._obtain_fs_observer)
    def _obtain_fs_observer(self) -> IFSObserver:
        return self._fs_observer

    @overrides(PythonLanguageServer._create_workspace)
    def _create_workspace(
        self, root_uri: str, fs_observer: IFSObserver, workspace_folders
    ):
        from robotframework_ls.impl.robot_workspace import RobotWorkspace

        return RobotWorkspace(
            root_uri, fs_observer, workspace_folders, generate_ast=False
        )

    def m_initialize(
        self,
        processId=None,
        rootUri=None,
        rootPath=None,
        initializationOptions=None,
        workspaceFolders=None,
        **_kwargs,
    ) -> dict:
        # capabilities = _kwargs.get("capabilities", {})
        # text_document_capabilities = capabilities.get("textDocument", {})
        # document_symbol_capabilities = text_document_capabilities.get(
        #     "documentSymbol", {}
        # )
        # hierarchical_document_symbol_support = document_symbol_capabilities.get(
        #     "hierarchicalDocumentSymbolSupport", False
        # )
        # self._hierarchical_document_symbol_support = (
        #     hierarchical_document_symbol_support
        # )

        ret = PythonLanguageServer.m_initialize(
            self,
            processId=processId,
            rootUri=rootUri,
            rootPath=rootPath,
            initializationOptions=initializationOptions,
            workspaceFolders=workspaceFolders,
            **_kwargs,
        )

        initialization_options = initializationOptions
        if initialization_options:
            plugins_dir = initialization_options.get("pluginsDir")
            if isinstance(plugins_dir, str):
                if not os.path.isdir(plugins_dir):
                    log.critical(f"Expected: {plugins_dir} to be a directory.")
                else:
                    self._pm.load_plugins_from(Path(plugins_dir))

        return ret

    def forward_msg(self, msg: dict) -> None:
        method = msg["method"]
        self._endpoint.notify(method, msg["params"])

    @overrides(PythonLanguageServer.capabilities)
    def capabilities(self):
        from robocorp_ls_core.lsp import TextDocumentSyncKind
        from robotframework_ls.impl.semantic_tokens import TOKEN_TYPES, TOKEN_MODIFIERS
        from robotframework_ls import commands

        server_capabilities = {
            "codeActionProvider": False,
            "codeLensProvider": {"resolveProvider": True},
            "completionProvider": {"resolveProvider": True},  # Docs are lazily computed
            "documentFormattingProvider": True,
            "documentHighlightProvider": True,
            "documentRangeFormattingProvider": False,
            "documentSymbolProvider": True,
            "definitionProvider": True,
            "executeCommandProvider": {
                "commands": [
                    "robot.addPluginsDir",
                    "robot.resolveInterpreter",
                    "robot.getLanguageServerVersion",
                    "robot.getInternalInfo",
                    "robot.listTests",
                ]
                + commands.ALL_SERVER_COMMANDS
            },
            "hoverProvider": True,
            "referencesProvider": True,
            "renameProvider": False,
            "foldingRangeProvider": True,
            # Note that there are no auto-trigger characters (there's no good
            # character as there's no `(` for parameters and putting it as a
            # space becomes a bit too much).
            # Under review: added ' ' and '\t' now that we have info on the
            # active parameter (waiting for feedback from community here to
            # check whether this is good or annoying).
            "signatureHelpProvider": {"triggerCharacters": [" ", "\t"]},
            "textDocumentSync": {
                "change": TextDocumentSyncKind.INCREMENTAL,
                "save": {"includeText": False},
                "openClose": True,
            },
            "workspace": {
                "workspaceFolders": {"supported": True, "changeNotifications": True}
            },
            "workspaceSymbolProvider": True,
            # The one below isn't accepted by lsp4j (it's still in LSP 3.15.0).
            # "workspaceSymbolProvider": {"workDoneProgress": False},
            "semanticTokensProvider": {
                "legend": {
                    "tokenTypes": TOKEN_TYPES,
                    "tokenModifiers": TOKEN_MODIFIERS,
                },
                "range": False,
                "full": True,
            },
        }
        log.debug("Server capabilities: %s", server_capabilities)
        return server_capabilities

    def m_workspace__execute_command(self, command: str = "", arguments=()) -> Any:
        if command.startswith("robot.internal.rfinteractive."):
            return rf_interactive_integration.execute_command(
                command, self, self._rf_interpreters_manager, arguments
            )

        return command_dispatcher.dispatch(self, command, arguments)

    @command_dispatcher("robot.addPluginsDir")
    def _add_plugins_dir(self, *arguments):
        directory: str = arguments[0]
        assert os.path.isdir(directory), f"Expected: {directory} to be a directory."
        self._pm.load_plugins_from(Path(directory))
        return True

    @command_dispatcher(ROBOT_START_INDEXING_INTERNAL)
    def _start_indexing(self, *arguments):
        self._server_manager.get_regular_rf_api_client("")

    def m_robot__provide_evaluatable_expression(
        self, uri: str, position: PositionTypedDict
    ):
        rf_api_client = self._server_manager.get_others_api_client(uri)
        if rf_api_client is not None:
            func = partial(
                self._threaded_api_request,
                rf_api_client,
                "request_evaluatable_expression",
                doc_uri=uri,
                position=position,
            )
            func = require_monitor(func)
            return func

        log.info("Unable to provide evaluatable expression (no api available).")
        return None

    @command_dispatcher(ROBOT_WAIT_FULL_TEST_COLLECTION_INTERNAL)
    def _wait_for_full_test_collection(self, *arguments):
        rf_api_client = self._server_manager.get_regular_rf_api_client("")
        if rf_api_client is not None:
            func = partial(
                self._threaded_api_request_no_doc,
                rf_api_client,
                "request_wait_for_full_test_collection",
            )
            func = require_monitor(func)
            return func

        log.info("Unable to wait for first test collection (no api available).")
        return []

    @command_dispatcher("robot.getInternalInfo")
    def _get_internal_info(self, *arguments):
        in_memory_docs = []
        workspace = self.workspace
        if workspace:
            for doc in workspace.iter_documents():
                in_memory_docs.append({"uri": doc.uri})
        return {
            "settings": self.config.get_full_settings(),
            "inMemoryDocs": in_memory_docs,
            "processId": os.getpid(),
        }

    @command_dispatcher("robot.resolveInterpreter")
    def _resolve_interpreter(self, *arguments):
        try:
            from robocorp_ls_core import uris
            from robotframework_ls.ep_resolve_interpreter import (
                EPResolveInterpreter,
            )
            from robotframework_ls.ep_resolve_interpreter import IInterpreterInfo

            target_robot = arguments[0]

            if isinstance(target_robot, (list, tuple)):
                if len(target_robot) > 0:
                    target_robot = target_robot[0]

            robot_uri = uris.from_fs_path(target_robot)

            for ep in self._pm.get_implementations(EPResolveInterpreter):
                interpreter_info: IInterpreterInfo = (
                    ep.get_interpreter_info_for_doc_uri(robot_uri)
                )
                if interpreter_info is not None:
                    return {
                        "pythonExe": interpreter_info.get_python_exe(),
                        "environ": interpreter_info.get_environ(),
                        "additionalPythonpathEntries": interpreter_info.get_additional_pythonpath_entries(),
                    }
        except:
            log.exception(f"Error resolving interpreter. Args: {arguments}")

    @command_dispatcher("robot.getLanguageServerVersion")
    def _get_language_server_version(self, *arguments):
        return __version__

    @command_dispatcher(ROBOT_GET_RFLS_HOME_DIR)
    def _get_rfls_home_dir(self, *arguments):
        from robotframework_ls import robot_config

        return robot_config.get_robotframework_ls_home()

    @command_dispatcher("robot.listTests")
    def _list_tests(self, *arguments):
        doc_uri = arguments[0]["uri"]

        rf_api_client = self._server_manager.get_others_api_client(doc_uri)
        if rf_api_client is not None:
            func = partial(
                self._threaded_api_request,
                rf_api_client,
                "request_list_tests",
                doc_uri=doc_uri,
            )
            func = require_monitor(func)
            return func

        log.info("Unable to list tests (no api available).")
        return []

    @overrides(PythonLanguageServer.m_workspace__did_change_configuration)
    @log_and_silence_errors(log)
    def m_workspace__did_change_configuration(self, **kwargs) -> None:
        PythonLanguageServer.m_workspace__did_change_configuration(self, **kwargs)
        self._server_manager.set_config(self.config)

    # --- Methods to forward to the api

    @overrides(PythonLanguageServer.m_shutdown)
    @log_and_silence_errors(log)
    def m_shutdown(self, **kwargs):
        try:
            from robocorp_ls_core.remote_fs_observer_impl import RemoteFSObserver

            remote_observer = typing.cast(RemoteFSObserver, self._fs_observer)
            remote_observer.dispose()
        except Exception:
            log.exception("Error disposing RemoteFSObserver.")
        self._server_manager.shutdown()

        PythonLanguageServer.m_shutdown(self, **kwargs)

    @overrides(PythonLanguageServer.m_exit)
    @log_and_silence_errors(log)
    def m_exit(self, **kwargs):
        self._server_manager.exit()

        PythonLanguageServer.m_exit(self, **kwargs)

    def m_text_document__formatting(
        self, textDocument=None, options=None
    ) -> Optional[list]:
        doc_uri = textDocument["uri"]

        source_format_rf_api_client = self._server_manager.get_others_api_client(
            doc_uri
        )
        if source_format_rf_api_client is None:
            log.info("Unable to get API for source format.")
            return []

        message_matcher = source_format_rf_api_client.request_source_format(
            text_document=textDocument, options=options
        )
        if message_matcher is None:
            raise RuntimeError(
                "Error requesting code formatting (message_matcher==None)."
            )
        curtime = time.time()
        maxtime = curtime + DEFAULT_COMPLETIONS_TIMEOUT

        # i.e.: wait X seconds for the code format and bail out if we
        # can't get it.
        available_time = maxtime - time.time()
        if available_time <= 0:
            raise RuntimeError("Code formatting timed-out (available_time <= 0).")

        if message_matcher.event.wait(available_time):
            msg = message_matcher.msg
            if msg is not None:
                result = msg.get("result")
                if result:
                    return result
                else:
                    return []
        raise RuntimeError("Code formatting timed-out.")

    @overrides(PythonLanguageServer.m_text_document__did_close)
    def m_text_document__did_close(self, textDocument=None, **_kwargs):
        self._server_manager.forward(
            ("api", "lint", "others"),
            "textDocument/didClose",
            {"textDocument": textDocument},
        )
        PythonLanguageServer.m_text_document__did_close(
            self, textDocument=textDocument, **_kwargs
        )

    @overrides(PythonLanguageServer.m_text_document__did_open)
    def m_text_document__did_open(self, textDocument=None, **_kwargs):
        self._server_manager.forward(
            ("api", "lint", "others"),
            "textDocument/didOpen",
            {"textDocument": textDocument},
        )
        PythonLanguageServer.m_text_document__did_open(
            self, textDocument=textDocument, **_kwargs
        )

    @overrides(PythonLanguageServer.m_text_document__did_change)
    def m_text_document__did_change(
        self, contentChanges=None, textDocument=None, **_kwargs
    ):
        self._server_manager.forward(
            ("api", "lint", "others"),
            "textDocument/didChange",
            {"contentChanges": contentChanges, "textDocument": textDocument},
        )
        PythonLanguageServer.m_text_document__did_change(
            self, contentChanges=contentChanges, textDocument=textDocument, **_kwargs
        )

    @overrides(PythonLanguageServer.m_workspace__did_change_workspace_folders)
    def m_workspace__did_change_workspace_folders(self, event=None, **_kwargs):
        self._server_manager.forward(
            ("api", "lint", "others"),
            "workspace/didChangeWorkspaceFolders",
            {"event": event},
        )
        PythonLanguageServer.m_workspace__did_change_workspace_folders(
            self, event=event, **_kwargs
        )

    # --- Customized implementation

    @overrides(PythonLanguageServer.lint)
    def lint(self, doc_uri, is_saved, content_changes=None) -> None:
        # We want the timeout to be shorter if it was saved or if the user
        # typed a new line.
        timeout = _LINT_DEBOUNCE_IN_SECONDS_LOW
        if not is_saved:
            timeout = _LINT_DEBOUNCE_IN_SECONDS_HIGH
            for change in content_changes:
                try:
                    text = change.get("text", "")
                    if "\n" in text or "\r" in text:
                        timeout = _LINT_DEBOUNCE_IN_SECONDS_LOW
                        break
                except:
                    log.exception(
                        "Error computing lint timeout with: %s", content_changes
                    )
        self._lint_manager.schedule_lint(doc_uri, is_saved, timeout)

    @overrides(PythonLanguageServer.cancel_lint)
    def cancel_lint(self, doc_uri) -> None:
        self._lint_manager.cancel_lint(doc_uri)

    def m_completion_item__resolve(self, **params):
        completion_item: CompletionItemTypedDict = params
        return self._robot_framework_ls_completion_impl.resolve_completion_item(
            completion_item
        )

    def m_text_document__completion(self, **params):
        doc_uri = params["textDocument"]["uri"]
        # Note: 0-based
        line, col = params["position"]["line"], params["position"]["character"]
        return self._robot_framework_ls_completion_impl.text_document_completion(
            doc_uri, line, col
        )

    @log_and_silence_errors(log)
    def _threaded_api_request(
        self,
        rf_api_client: IRobotFrameworkApiClient,
        request_method_name: str,
        doc_uri: str,
        monitor: IMonitor,
        __timeout__=DEFAULT_COMPLETIONS_TIMEOUT,
        __log__=False,
        **kwargs,
    ):
        from robocorp_ls_core.client_base import wait_for_message_matcher

        func = getattr(rf_api_client, request_method_name)

        ws = self.workspace
        if not ws:
            log.critical(
                "Workspace must be set before calling %s.", request_method_name
            )
            return None

        document = ws.get_document(doc_uri, accept_from_file=True)
        if document is None:
            log.critical(
                "Unable to find document (%s) for %s." % (doc_uri, request_method_name)
            )
            return None

        # Asynchronous completion.
        message_matcher: Optional[IIdMessageMatcher] = func(doc_uri, **kwargs)
        if message_matcher is None:
            log.debug("Message matcher for %s returned None.", request_method_name)
            return None

        if wait_for_message_matcher(
            message_matcher,
            rf_api_client.request_cancel,
            __timeout__,
            monitor,
        ):
            msg = message_matcher.msg
            if msg is not None:
                result = msg.get("result")
                if __log__:
                    log.info("Result: %s", result)
                if result:
                    return result

        return None

    @log_and_silence_errors(log)
    def _threaded_api_request_no_doc(
        self,
        rf_api_client: IRobotFrameworkApiClient,
        request_method_name: str,
        monitor: Optional[IMonitor],
        **kwargs,
    ):
        from robocorp_ls_core.client_base import wait_for_message_matcher

        func = getattr(rf_api_client, request_method_name)

        # Asynchronous completion.
        message_matcher: Optional[IIdMessageMatcher] = func(**kwargs)
        if message_matcher is None:
            log.debug("Message matcher for %s returned None.", request_method_name)
            return None

        if wait_for_message_matcher(
            message_matcher,
            rf_api_client.request_cancel,
            DEFAULT_COMPLETIONS_TIMEOUT,
            monitor,
        ):
            msg = message_matcher.msg
            if msg is not None:
                result = msg.get("result")
                if result:
                    return result

        return None

    def async_api_forward(
        self,
        api_client_method_name: str,
        target_api: str,  # api, lint, others
        doc_uri: str,
        default_return=None,
        __add_doc_uri_in_args__=True,
        **kwargs,
    ):
        rf_api_client: Optional[IRobotFrameworkApiClient]
        if target_api == "api":
            rf_api_client = self._server_manager.get_regular_rf_api_client(doc_uri)
        elif target_api == "lint":
            rf_api_client = self._server_manager.get_lint_rf_api_client(doc_uri)
        elif target_api == "others":
            rf_api_client = self._server_manager.get_others_api_client(doc_uri)
        else:
            raise AssertionError(f"Unexpected: {target_api}")

        self._last_doc_uri = doc_uri
        if rf_api_client is not None:
            if __add_doc_uri_in_args__:
                func = partial(
                    self._threaded_api_request,
                    rf_api_client,
                    api_client_method_name,
                    doc_uri=doc_uri,
                    **kwargs,
                )
            else:
                func = partial(
                    self._threaded_api_request_no_doc,
                    rf_api_client,
                    api_client_method_name,
                    **kwargs,
                )
            func = require_monitor(func)
            return func

        log.info(
            "No api available (call: %s, uri: %s).", api_client_method_name, doc_uri
        )
        return default_return

    def m_text_document__definition(self, **kwargs):
        doc_uri = kwargs["textDocument"]["uri"]
        # Note: 0-based
        line, col = kwargs["position"]["line"], kwargs["position"]["character"]
        return self.async_api_forward(
            "request_find_definition", "api", doc_uri, line=line, col=col
        )

    def m_text_document__signature_help(self, **kwargs):
        """
        "params": {
            "textDocument": {
                "uri": "file:///x%3A/vscode-robot/local_test/Basic/resources/keywords.robot"
            },
            "position": {"line": 7, "character": 22},
            "context": {
                "isRetrigger": False,
                "triggerCharacter": " ",
                "triggerKind": 2,
            },
        },
        """
        doc_uri = kwargs["textDocument"]["uri"]
        # Note: 0-based
        line, col = kwargs["position"]["line"], kwargs["position"]["character"]

        return self.async_api_forward(
            "request_signature_help", "api", doc_uri, line=line, col=col
        )

    def m_text_document__folding_range(self, **kwargs):
        """
        "params": {
            "textDocument": {
                "uri": "file:///x%3A/vscode-robot/local_test/Basic/resources/keywords.robot"
            },
        },
        """
        doc_uri = kwargs["textDocument"]["uri"]

        return self.async_api_forward("request_folding_range", "others", doc_uri)

    def m_text_document__code_lens(self, **kwargs):
        doc_uri = kwargs["textDocument"]["uri"]
        return self.async_api_forward("request_code_lens", "others", doc_uri)

    def m_code_lens__resolve(self, **kwargs):
        code_lens: CodeLensTypedDict = kwargs

        code_lens_command = code_lens.get("command")
        data = code_lens.get("data")
        if code_lens_command is None and isinstance(data, dict):
            # For the interactive shell we need to resolve the arguments.
            doc_uri = data.get("uri")

            return self.async_api_forward(
                "request_resolve_code_lens",
                "others",
                doc_uri,
                code_lens=code_lens,
                __add_doc_uri_in_args__=False,
            )

        return code_lens

    def m_text_document__document_symbol(self, **kwargs):
        doc_uri = kwargs["textDocument"]["uri"]
        return self.async_api_forward("request_document_symbol", "others", doc_uri)

    def m_text_document__hover(self, **kwargs):
        params: TextDocumentPositionParamsTypedDict = kwargs
        doc_uri = params["textDocument"]["uri"]
        # Note: 0-based
        line, col = params["position"]["line"], params["position"]["character"]

        return self.async_api_forward(
            "request_hover", "api", doc_uri, line=line, col=col
        )

    def m_text_document__references(self, **kwargs):
        doc_uri = kwargs["textDocument"]["uri"]
        # Note: 0-based
        line, col = kwargs["position"]["line"], kwargs["position"]["character"]
        include_declaration = kwargs["context"]["includeDeclaration"]

        # Note: we want to use the same one used by m_workspace__symbol (to reuse
        # the related caches).
        rf_api_client = self._server_manager.get_regular_rf_api_client(doc_uri)
        if rf_api_client is not None:
            func = partial(
                self._threaded_api_request,
                rf_api_client,
                "request_references",
                doc_uri=doc_uri,
                line=line,
                col=col,
                include_declaration=include_declaration,
                __timeout__=9999999,
            )
            func = require_monitor(func)
            return func

        log.info("Unable to compute references (no api available).")
        return []

    def m_text_document__semantic_tokens__range(self, textDocument=None, range=None):
        raise RuntimeError("Not currently implemented!")

    def m_text_document__semantic_tokens__full(self, textDocument=None):
        doc_uri = textDocument["uri"]

        return self.async_api_forward(
            "request_semantic_tokens_full",
            "others",
            doc_uri,
            default_return={"resultId": None, "data": []},
            text_document=textDocument,
            __add_doc_uri_in_args__=False,
        )

    def m_workspace__symbol(self, query: Optional[str] = None) -> Any:
        doc_uri = self._last_doc_uri
        return self.async_api_forward(
            "request_workspace_symbols",
            "api",
            doc_uri,
            query=query,
            __add_doc_uri_in_args__=False,
        )

    def m_text_document__document_highlight(self, **kwargs):
        params: TextDocumentPositionParamsTypedDict = kwargs
        doc_uri = params["textDocument"]["uri"]
        line, col = params["position"]["line"], params["position"]["character"]

        return self.async_api_forward(
            "request_document_highlight", "others", doc_uri, line=line, col=col
        )



from _pydevd_bundle.pydevd_constants import DebugInfoHolder, SHOW_COMPILE_CYTHON_COMMAND_LINE, NULL, LOG_TIME
from contextlib import contextmanager
import traceback
import os
import sys


class _LoggingGlobals(object):

    _warn_once_map = {}
    _debug_stream_filename = None
    _debug_stream = sys.stderr
    _debug_stream_initialized = False


def initialize_debug_stream(reinitialize=False):
    '''
    :param bool reinitialize:
        Reinitialize is used to update the debug stream after a fork (thus, if it wasn't
        initialized, we don't need to do anything).
    '''
    if reinitialize:
        if not _LoggingGlobals._debug_stream_initialized:
            return
    else:
        if _LoggingGlobals._debug_stream_initialized:
            return

    _LoggingGlobals._debug_stream_initialized = True

    # Note: we cannot initialize with sys.stderr because when forking we may end up logging things in 'os' calls.
    _LoggingGlobals._debug_stream = NULL
    _LoggingGlobals._debug_stream_filename = None

    if not DebugInfoHolder.PYDEVD_DEBUG_FILE:
        _LoggingGlobals._debug_stream = sys.stderr
    else:
        # Add pid to the filename.
        try:
            dirname = os.path.dirname(DebugInfoHolder.PYDEVD_DEBUG_FILE)
            basename = os.path.basename(DebugInfoHolder.PYDEVD_DEBUG_FILE)
            try:
                os.makedirs(dirname)
            except:
                pass  # Ignore error if it already exists.

            name, ext = os.path.splitext(basename)
            debug_file = os.path.join(dirname, name + '.' + str(os.getpid()) + ext)
            _LoggingGlobals._debug_stream = open(debug_file, 'w')
            _LoggingGlobals._debug_stream_filename = debug_file
        except:
            _LoggingGlobals._debug_stream = sys.stderr
            # Don't fail when trying to setup logging, just show the exception.
            traceback.print_exc()


def list_log_files(pydevd_debug_file):
    log_files = []
    dirname = os.path.dirname(pydevd_debug_file)
    basename = os.path.basename(pydevd_debug_file)
    if os.path.isdir(dirname):
        name, ext = os.path.splitext(basename)
        for f in os.listdir(dirname):
            if f.startswith(name) and f.endswith(ext):
                log_files.append(os.path.join(dirname, f))
    return log_files


@contextmanager
def log_context(trace_level, stream):
    '''
    To be used to temporarily change the logging settings.
    '''
    original_trace_level = DebugInfoHolder.DEBUG_TRACE_LEVEL
    original_debug_stream = _LoggingGlobals._debug_stream
    original_pydevd_debug_file = DebugInfoHolder.PYDEVD_DEBUG_FILE
    original_debug_stream_filename = _LoggingGlobals._debug_stream_filename
    original_initialized = _LoggingGlobals._debug_stream_initialized

    DebugInfoHolder.DEBUG_TRACE_LEVEL = trace_level
    _LoggingGlobals._debug_stream = stream
    _LoggingGlobals._debug_stream_initialized = True
    try:
        yield
    finally:
        DebugInfoHolder.DEBUG_TRACE_LEVEL = original_trace_level
        _LoggingGlobals._debug_stream = original_debug_stream
        DebugInfoHolder.PYDEVD_DEBUG_FILE = original_pydevd_debug_file
        _LoggingGlobals._debug_stream_filename = original_debug_stream_filename
        _LoggingGlobals._debug_stream_initialized = original_initialized


import time
_last_log_time = time.time()


def _pydevd_log(level, msg, *args):
    '''
    Levels are:

    0 most serious warnings/errors (always printed)
    1 warnings/significant events
    2 informational trace
    3 verbose mode
    '''
    if level <= DebugInfoHolder.DEBUG_TRACE_LEVEL:
        # yes, we can have errors printing if the console of the program has been finished (and we're still trying to print something)
        try:
            try:
                if args:
                    msg = msg % args
            except:
                msg = '%s - %s' % (msg, args)

            if LOG_TIME:
                global _last_log_time
                new_log_time = time.time()
                time_diff = new_log_time - _last_log_time
                _last_log_time = new_log_time
                msg = '%.2fs - %s\n' % (time_diff, msg,)
            else:
                msg = '%s\n' % (msg,)
            try:
                try:
                    initialize_debug_stream()  # Do it as late as possible
                    _LoggingGlobals._debug_stream.write(msg)
                except TypeError:
                    if isinstance(msg, bytes):
                        # Depending on the StringIO flavor, it may only accept unicode.
                        msg = msg.decode('utf-8', 'replace')
                        _LoggingGlobals._debug_stream.write(msg)
            except UnicodeEncodeError:
                # When writing to the stream it's possible that the string can't be represented
                # in the encoding expected (in this case, convert it to the stream encoding
                # or ascii if we can't find one suitable using a suitable replace).
                encoding = getattr(_LoggingGlobals._debug_stream, 'encoding', 'ascii')
                msg = msg.encode(encoding, 'backslashreplace')
                msg = msg.decode(encoding)
                _LoggingGlobals._debug_stream.write(msg)

            _LoggingGlobals._debug_stream.flush()
        except:
            pass
        return True


def _pydevd_log_exception(msg='', *args):
    if msg or args:
        _pydevd_log(0, msg, *args)
    try:
        initialize_debug_stream()  # Do it as late as possible
        traceback.print_exc(file=_LoggingGlobals._debug_stream)
        _LoggingGlobals._debug_stream.flush()
    except:
        raise


def verbose(msg, *args):
    if DebugInfoHolder.DEBUG_TRACE_LEVEL >= 3:
        _pydevd_log(3, msg, *args)


def debug(msg, *args):
    if DebugInfoHolder.DEBUG_TRACE_LEVEL >= 2:
        _pydevd_log(2, msg, *args)


def info(msg, *args):
    if DebugInfoHolder.DEBUG_TRACE_LEVEL >= 1:
        _pydevd_log(1, msg, *args)


warn = info


def critical(msg, *args):
    _pydevd_log(0, msg, *args)


def exception(msg='', *args):
    try:
        _pydevd_log_exception(msg, *args)
    except:
        pass  # Should never fail (even at interpreter shutdown).


error = exception


def error_once(msg, *args):
    try:
        if args:
            message = msg % args
        else:
            message = str(msg)
    except:
        message = '%s - %s' % (msg, args)

    if message not in _LoggingGlobals._warn_once_map:
        _LoggingGlobals._warn_once_map[message] = True
        critical(message)


def exception_once(msg, *args):
    try:
        if args:
            message = msg % args
        else:
            message = str(msg)
    except:
        message = '%s - %s' % (msg, args)

    if message not in _LoggingGlobals._warn_once_map:
        _LoggingGlobals._warn_once_map[message] = True
        exception(message)


def debug_once(msg, *args):
    if DebugInfoHolder.DEBUG_TRACE_LEVEL >= 3:
        error_once(msg, *args)


def show_compile_cython_command_line():
    if SHOW_COMPILE_CYTHON_COMMAND_LINE:
        dirname = os.path.dirname(os.path.dirname(__file__))
        error_once("warning: Debugger speedups using cython not found. Run '\"%s\" \"%s\" build_ext --inplace' to build.",
            sys.executable, os.path.join(dirname, 'setup_pydevd_cython.py'))




import sys
import threading
import traceback


def dump_threads():
    thread_id_to_name = {}
    try:
        for t in threading.enumerate():
            thread_id_to_name[t.ident] = "%s  (daemon: %s)" % (t.name, t.daemon)
    except:
        pass

    stack_trace = [
        "===============================================================================",
        "Threads still found running after tests finished",
        "================================= Thread Dump =================================",
    ]

    for thread_id, stack in sys._current_frames().items():
        stack_trace.append(
            "\n-------------------------------------------------------------------------------"
        )
        stack_trace.append(" Thread %s" % thread_id_to_name.get(thread_id, thread_id))
        stack_trace.append("")

        if "self" in stack.f_locals:
            sys.stderr.write(str(stack.f_locals["self"]) + "\n")

        for filename, lineno, name, line in traceback.extract_stack(stack):
            stack_trace.append(' File "%s", line %d, in %s' % (filename, lineno, name))
            if line:
                stack_trace.append("   %s" % (line.strip()))
    stack_trace.append(
        "\n=============================== END Thread Dump ==============================="
    )
    sys.stderr.write("\n".join(stack_trace))



# TODO: this is not the best e.g. we capture numbers
import re

from robocorp_ls_core.protocols import IDocumentSelection, IDocument


RE_START_WORD = re.compile("[\w]*$")
RE_END_WORD = re.compile("^[\w]*")


class DocumentSelection(object):
    def __init__(self, doc: IDocument, line: int, col: int):
        if line < 0:
            line = 0

        if col < 0:
            col = 0

        self.doc = doc
        self.line = line
        self.col = col

    @property
    def offset_at_position(self):
        """Return the byte-offset pointed at by the given position."""
        offset = 0
        for i, curr_line in enumerate(self.doc.iter_lines()):
            if i == self.line:
                break
            offset += len(curr_line)

        return offset + self.col

    @property
    def current_line(self) -> str:
        return self.doc.get_line(self.line)

    @property
    def line_to_column(self):
        current_line = self.current_line
        if not current_line:
            return ""
        line_start = current_line[: self.col]

        return line_start

    @property
    def word_at_column(self) -> str:
        current_line = self.current_line
        if not current_line:
            return ""

        col = self.col
        # Split word in two
        start = current_line[:col]
        end = current_line[col:]

        # Take end of start and start of end to find word
        # These are guaranteed to match, even if they match the empty string
        m_start = RE_START_WORD.findall(start)
        m_end = RE_END_WORD.findall(end)

        return m_start[0] + m_end[-1]

    @property
    def word_to_column(self) -> str:
        line_to_cursor = self.line_to_column

        m_start = RE_START_WORD.findall(line_to_cursor)

        return m_start[0]

    @property
    def word_from_column(self) -> str:
        current_line = self.current_line
        if not current_line:
            return ""

        col = self.col
        # Split word in two
        end = current_line[col:]

        m_end = RE_END_WORD.findall(end)

        return m_end[-1]

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements

        _: IDocumentSelection = check_implements(self)



from robotframework_ls.impl.text_utilities import get_digest_from_string
import os
from typing import Optional
from robotframework_ls.impl.protocols import ILibraryDoc
from robocorp_ls_core.robotframework_log import get_logger
import threading
import robotframework_ls
import robocorp_ls_core
import weakref
from robocorp_ls_core.basic import normalize_filename
import sys


log = get_logger(__name__)


def _get_mtime_from_stream(stream) -> Optional[float]:
    lst = []
    while True:
        c = stream.read(1)
        if c == "\n":
            break
        lst.append(c)
    line = "".join(lst).strip()

    if not line.startswith("mtime:"):
        log.info(
            "Expected %s contents to start with 'mtime:<saved_mtime>'\\n. Found: %s",
            target_json,
            line,
        )
        return None

    try:
        return float(line.split(":")[1])
    except:
        log.info(
            "Unable to load from json: %s because mtime line (%s) is not what we expected.",
            target_json,
            line,
        )
        return None


def load_markdown_json_version(
    libspec_manager, spec_filename, mtime: float
) -> Optional[ILibraryDoc]:
    from robotframework_ls.impl import robot_specbuilder

    target_json = _get_markdown_json_version_filename(libspec_manager, spec_filename)

    stream = None
    try:
        try:
            stream = open(target_json, "r", encoding="utf-8")
        except:
            log.debug("Unable to load from json: %s (file does not exist)", target_json)
        else:
            try:
                loaded_mtime = _get_mtime_from_stream(stream)

                if str(loaded_mtime) != str(mtime):
                    log.debug(
                        "Unable to load from json: %s because mtime no longer matches.",
                        target_json,
                    )
                    return None

                # Note: let the external world think it was built with the libspec
                # (so, the stream is from the json and the spec filename is from .libspec).
                json_builder = robot_specbuilder.JsonDocBuilder()
                return json_builder.build_from_stream(spec_filename, stream)
            except:
                log.exception("Error loading libdoc from json: %s", target_json)
                return None

    finally:
        if stream is not None:
            stream.close()

    return None


def _convert_to_markdown_if_needed(spec_filename, target_json) -> None:
    import tempfile
    import shutil

    try:
        with open(target_json, "r", encoding="utf-8") as existing_stream:
            current_mtime = _get_mtime_from_stream(existing_stream)
    except:
        current_mtime = None

    from robotframework_ls.impl import robot_specbuilder

    # If it raises because it's not there, that's ok!
    mtime = os.path.getmtime(spec_filename)

    if str(mtime) == str(current_mtime):
        # No need to convert, it still matches.
        return

    builder = robot_specbuilder.SpecDocBuilder()
    libdoc = builder.build(spec_filename)
    if libdoc.doc_format == "markdown":
        # i.e.: it's already in markdown.
        return

    libdoc.convert_docs_to_markdown()

    try:
        os.makedirs(os.path.dirname(target_json), exist_ok=True)
    except:
        pass

    with tempfile.NamedTemporaryFile(
        mode="w+", dir=os.path.dirname(target_json), delete=False, encoding="utf-8"
    ) as tempf:
        import json

        tempf.write(f"mtime:{mtime}\n")

        json.dump(libdoc.to_dictionary(), tempf)

    shutil.copy(tempf.name, target_json)
    os.remove(tempf.name)


class _ConversionThread(threading.Thread):

    DISPOSE = "DISPOSE"

    def __init__(self):
        import queue

        threading.Thread.__init__(self)
        self.daemon = True
        self.queue = queue.Queue()

    def run(self):
        from robocorp_ls_core.subprocess_wrapper import subprocess

        while True:
            entry = self.queue.get()
            if entry == self.DISPOSE:
                return

            spec_filename, target_json = entry
            env = os.environ.copy()

            # Make sure we're in the pythonpath.
            env["PYTHONPATH"] = os.pathsep.join(
                [
                    os.path.dirname(os.path.dirname(robotframework_ls.__file__)),
                    os.path.dirname(os.path.dirname(robocorp_ls_core.__file__)),
                ]
            )

            try:
                subprocess.check_output(
                    [sys.executable, __file__, spec_filename, target_json],
                    stderr=subprocess.STDOUT,
                    stdin=subprocess.PIPE,
                    env=env,
                )
            except subprocess.CalledProcessError as e:
                log.exception(
                    "Error converting libspec to markdown: %s.\nReturn code: %s\nOutput:\n%s",
                    spec_filename,
                    e.returncode,
                    e.output,
                )


def _get_markdown_json_version_filename(libspec_manager, spec_filename: str) -> str:
    spec_filename = normalize_filename(spec_filename)
    digest = get_digest_from_string(spec_filename)

    target_json = os.path.join(
        libspec_manager.cache_libspec_dir,
        f"{digest}_{os.path.basename(spec_filename)}.json",
    )
    return target_json


class LibspecMarkdownConversion:
    def __init__(self, libspec_manager):
        self._conversion_thread = _ConversionThread()
        self._started = False
        self._weak_libspec_manager = weakref.ref(libspec_manager)

    def get_markdown_json_version_filename(self, spec_filename: str) -> str:
        libspec_manager = self._weak_libspec_manager()
        assert libspec_manager is not None
        return _get_markdown_json_version_filename(libspec_manager, spec_filename)

    def schedule_conversion_to_markdown(self, spec_filename: str) -> Optional[str]:
        if not self._started:
            self._conversion_thread.start()
            self._started = True

        libspec_manager = self._weak_libspec_manager()
        if libspec_manager is None:
            self.dispose()
            return None

        target_json = self.get_markdown_json_version_filename(spec_filename)

        self._conversion_thread.queue.put((spec_filename, target_json))
        return target_json

    def dispose(self):
        self._conversion_thread.queue.put(_ConversionThread.DISPOSE)


if __name__ == "__main__":
    args = sys.argv[1:]

    try:
        spec_filename = args[0]
        target_json = args[1]
    except:
        sys.stderr.write(
            f"Expected 2 arguments (spec_filename, target_json). Received: {args}"
        )
        sys.stderr.flush()
        sys.exit(1)

    _convert_to_markdown_if_needed(spec_filename, target_json)



import re
import ast
from collections import Counter

from robot.api.parsing import ModelTransformer, Variable, Token
from robotidy.exceptions import InvalidParameterValueError


class NormalizeAssignments(ModelTransformer):
    """
    Normalize assignments.

    It can change all assignment signs to either the most commonly used in a given file or a configured fixed one.
    Default behaviour is autodetect for assignments from Keyword Calls and removing assignment signs in
    *** Variables *** section. It can be freely configured.

    In this code most common is no equal sign at all. We should remove `=` signs from all lines:

        *** Variables ***
        ${var} =  ${1}
        @{list}  a
        ...  b
        ...  c

        ${variable}=  10


        *** Keywords ***
        Keyword
            ${var}  Keyword1
            ${var}   Keyword2
            ${var}=    Keyword

    To:

        *** Variables ***
        ${var}  ${1}
        @{list}  a
        ...  b
        ...  c

        ${variable}  10


        *** Keywords ***
        Keyword
            ${var}  Keyword1
            ${var}   Keyword2
            ${var}    Keyword

    You can configure that behaviour to automatically add desired equal sign with `equal_sign_type`
    (default `autodetect`) and `equal_sign_type_variables` (default `remove`) parameters.
    (possible types are: `autodetect`, `remove`, `equal_sign` ('='), `space_and_equal_sign` (' =').

    See https://robotidy.readthedocs.io/en/latest/transformers/NormalizeAssignments.html for more examples.
    """

    def __init__(
        self,
        equal_sign_type: str = "autodetect",
        equal_sign_type_variables: str = "remove",
    ):
        self.remove_equal_sign = re.compile(r"\s?=$")
        self.file_equal_sign_type = None
        self.file_equal_sign_type_variables = None
        self.equal_sign_type = self.parse_equal_sign_type(equal_sign_type, "equal_sign_type")
        self.equal_sign_type_variables = self.parse_equal_sign_type(
            equal_sign_type_variables, "equal_sign_type_variables"
        )

    def parse_equal_sign_type(self, value, name):
        types = {
            "remove": "",
            "equal_sign": "=",
            "space_and_equal_sign": " =",
            "autodetect": None,
        }
        if value not in types:
            raise InvalidParameterValueError(
                self.__class__.__name__,
                value,
                name,
                "Possible values:\n    remove\n    equal_sign\n    space_and_equal_sign",
            )
        return types[value]

    def visit_File(self, node):  # noqa
        """
        If no assignment sign was set the file will be scanned to find most common assignment sign.
        This auto detection will happen for every file separately.
        """
        if self.equal_sign_type is None or self.equal_sign_type_variables is None:
            common, common_variables = self.auto_detect_equal_sign(node)
            if self.equal_sign_type is None and common is not None:
                self.file_equal_sign_type = common
            if self.equal_sign_type_variables is None and common_variables is not None:
                self.file_equal_sign_type_variables = common_variables
            if self.file_equal_sign_type is None and self.file_equal_sign_type_variables is None:
                return node
        self.generic_visit(node)
        self.file_equal_sign_type = None
        self.file_equal_sign_type_variables = None

    def visit_KeywordCall(self, node):  # noqa
        if node.assign:  # if keyword returns any value
            assign_tokens = node.get_tokens(Token.ASSIGN)
            self.normalize_equal_sign(assign_tokens[-1], self.equal_sign_type, self.file_equal_sign_type)
        return node

    def visit_VariableSection(self, node):  # noqa
        for child in node.body:
            if not isinstance(child, Variable):
                continue
            var_token = child.get_token(Token.VARIABLE)
            self.normalize_equal_sign(
                var_token,
                self.equal_sign_type_variables,
                self.file_equal_sign_type_variables,
            )
        return node

    def normalize_equal_sign(self, token, overwrite, local_normalize):
        token.value = re.sub(self.remove_equal_sign, "", token.value)
        if overwrite:
            token.value += overwrite
        elif local_normalize:
            token.value += local_normalize

    @staticmethod
    def auto_detect_equal_sign(node):
        auto_detector = AssignmentTypeDetector()
        auto_detector.visit(node)
        return auto_detector.most_common, auto_detector.most_common_variables


class AssignmentTypeDetector(ast.NodeVisitor):
    def __init__(self):
        self.sign_counter, self.sign_counter_variables = Counter(), Counter()
        self.most_common = None
        self.most_common_variables = None

    def visit_File(self, node):  # noqa
        self.generic_visit(node)
        if len(self.sign_counter) >= 2:
            self.most_common = self.sign_counter.most_common(1)[0][0]
        if len(self.sign_counter_variables) >= 2:
            self.most_common_variables = self.sign_counter_variables.most_common(1)[0][0]

    def visit_KeywordCall(self, node):  # noqa
        if node.assign:  # if keyword returns any value
            sign = self.get_assignment_sign(node.assign[-1])
            self.sign_counter[sign] += 1

    def visit_VariableSection(self, node):  # noqa
        for child in node.body:
            if not isinstance(child, Variable):
                continue
            var_token = child.get_token(Token.VARIABLE)
            sign = self.get_assignment_sign(var_token.value)
            self.sign_counter_variables[sign] += 1
        return node

    @staticmethod
    def get_assignment_sign(token_value):
        return token_value[token_value.find("}") + 1 :]



from robocorp_ls_core.lsp import (
    HoverTypedDict,
    MarkupKind,
    SignatureHelp,
    SignatureInformation,
    MarkupContentTypedDict,
)
from typing import Optional


def hover(completion_context) -> Optional[HoverTypedDict]:
    from robotframework_ls.impl.find_definition import find_definition_extended
    from robotframework_ls.impl import ast_utils

    definition_info = find_definition_extended(completion_context)
    if definition_info:
        for definition in definition_info.definitions:
            if hasattr(definition, "keyword_found"):
                # If we found a keyword use the signature help.
                break

            return {
                "contents": definition.hover_docs(),
                "range": definition_info.origin_selection_range,
            }

    from robotframework_ls.impl.signature_help import signature_help_internal

    sig_help: Optional[SignatureHelp] = signature_help_internal(completion_context)
    if sig_help is None:
        return None

    node = sig_help.node
    signatures = sig_help.signatures
    if not signatures:
        return None
    try:
        active_signature: SignatureInformation = signatures[sig_help.activeSignature]
    except IndexError:
        active_signature = signatures[0]

    active_parameter = sig_help.activeParameter

    optional_documentation_markup: Optional[
        MarkupContentTypedDict
    ] = active_signature.documentation
    documentation_markup: MarkupContentTypedDict
    if not optional_documentation_markup:
        documentation_markup = {"kind": MarkupKind.Markdown, "value": ""}
    else:
        documentation_markup = optional_documentation_markup
    kind = documentation_markup["kind"]

    # Now, let's add the signature to the documentation
    escape = lambda s: s

    if kind == MarkupKind.Markdown:
        from robotframework_ls import html_to_markdown

        escape = html_to_markdown.escape

    add_documentation = True
    if kind == MarkupKind.Markdown:
        signature_doc = ["**", escape(sig_help.name), "**"]
    else:
        signature_doc = [sig_help.name]

    if kind == MarkupKind.Markdown:
        prefix_highlight = "*`"
        postfix_highlight = "`*"
    else:
        prefix_highlight = "`"
        postfix_highlight = "`"

    if active_signature.parameters:
        signature_doc.append("(")
        for i, parameter in enumerate(active_signature.parameters):
            if i > 0:
                signature_doc.append(", ")

            escaped_label = escape(parameter.label)
            if i == active_parameter:
                add_documentation = False
                signature_doc.insert(
                    0,
                    f"Parameter: {prefix_highlight}{escaped_label}{postfix_highlight} in Keyword Call.\n\n",
                )

                signature_doc.append(prefix_highlight)

            signature_doc.append(escaped_label)

            if i == active_parameter:
                signature_doc.append(postfix_highlight)

        signature_doc.append(")")

    if add_documentation:
        # When over a parameter, don't add the documentation.
        signature_doc.append("\n\n")
        signature_doc.append(documentation_markup["value"])

    token_info = completion_context.get_current_token()
    if token_info and token_info.token:
        show_range = ast_utils.create_range_from_token(token_info.token)
    else:
        show_range = {
            "start": {"line": node.lineno - 1, "character": node.col_offset},
            "end": {"line": node.end_lineno - 1, "character": node.end_col_offset},
        }
    return {
        "contents": {"kind": kind, "value": "".join(signature_doc)},
        "range": show_range,
    }



import pytest
import sys
from _pydevd_bundle.pydevd_constants import NO_FTRACE
from tests_python.debugger_unittest import IS_JYTHON


class Tracer(object):

    def __init__(self):
        self.call_calls = 0
        self.line_calls = 0

    def trace_func(self, frame, event, arg):
        if event == 'call':
            self.call_calls += 1
            return self.on_call()

        elif event == 'line':
            self.line_calls += 1
            # Should disable tracing when None is returned (but doesn't).
            return self.on_line(frame, event, arg)

        else:
            return self.trace_func

    def on_call(self):
        return self.trace_func

    def on_line(self, frame, event, arg):
        return None


class TracerSettingNone(Tracer):

    def on_line(self, frame, event, arg):
        frame.f_trace = NO_FTRACE
        return NO_FTRACE


class TracerChangeToOtherTracing(Tracer):

    def on_line(self, frame, event, arg):
        return self._no_trace

    def _no_trace(self, frame, event, arg):
        return self._no_trace


class TracerDisableOnCall(Tracer):

    def on_call(self):
        return None


def test_tracing_gotchas():
    '''
    Summary of the gotchas tested:

    If 'call' is used, a None return value is used for the tracing. Afterwards the return may or may
    not be ignored depending on the Python version.

    Also, on Python 2.6, the trace function may not be set to None as it'll give an error
    afterwards (it needs to be set to an empty tracing function).

    All of the versions seem to work when the return value is a new callable though.

    Note: according to: https://docs.python.org/3/library/sys.html#sys.settrace the behavior
    does not follow the spec (but we have to work with it nonetheless).

    Note: Jython seems to do what's written in the docs.
    '''

    def method():
        _a = 1
        _b = 1
        _c = 1
        _d = 1

    for tracer, line_events in (
        (Tracer(), 1 if IS_JYTHON else 4),
        (TracerSettingNone(), 1),
        (TracerChangeToOtherTracing(), 1),
        (TracerDisableOnCall(), 0),
        ):
        curr_trace_func = sys.gettrace()
        try:
            sys.settrace(tracer.trace_func)

            method()

            if tracer.call_calls != 1:
                pytest.fail('Expected a single call event. Found: %s' % (tracer.call_calls))

            if tracer.line_calls != line_events:
                pytest.fail('Expected %s line events. Found: %s. Tracer: %s' % (line_events, tracer.line_calls, tracer))
        finally:
            sys.settrace(curr_trace_func)


if __name__ == '__main__':
    pytest.main(['-k', 'test_tracing_gotchas'])
